import logging
import os
from pathlib import Path

import numpy as np
import torch
import torch.nn as nn
from kliff.descriptors.descriptor import Descriptor
from kliff.utils import create_directory, seed_all, to_path

logger = logging.getLogger(__name__)


class ModelTorch(nn.Module):
    """
    Base class for machine learning models.

    Typically, a user will not directly use this.

    Args:
        descriptor: atomic environment descriptor for computing configuration
            fingerprints. See :meth:`~kliff.descriptors.SymmetryFunction` and
            :meth:`~kliff.descriptors.Bispectrum`.
        seed: random seed.
    """

    def __init__(self, descriptor: Descriptor, seed: int = 35):
        super(ModelTorch, self).__init__()

        self.descriptor = descriptor
        self.seed = seed
        seed_all(seed)

        dtype = self.descriptor.get_dtype()

        if dtype == np.float32:
            self.dtype = torch.float32
        elif dtype == np.float64:
            self.dtype = torch.float64
        else:
            raise ModelTorchError(f"Not support dtype {dtype}.")

        self.save_prefix = None
        self.save_start = None
        self.save_frequency = None

    def forward(self, x):
        raise NotImplementedError("`forward` not implemented.")

    def fit(self, path: Path):
        """
        Fit the model using analytic solution.

        Args:
            path: path to the fingerprints generated by the descriptor.
        """
        raise ModelTorchError(
            "Analytic fitting not supported for this model. Minimize a loss function to "
            "train the model instead."
        )

    def set_save_metadata(self, prefix: Path, start: int, frequency: int = 1):
        """
        Set metadata that controls how the model are saved during training.

        If this function is called before minimization starts, the model will be
        saved to the directory specified by `prefix` every `frequency` epochs,
        beginning at the `start` epoch.

        Args:
            prefix: Path to the directory where the models are saved.
                Model will be named as `{prefix}/model_epoch{ep}.pt`, where `ep` is
                the epoch number.
            start: Epoch number at which begins to save the model.
            frequency: Save the model every `frequency` epochs.
        """
        self.save_prefix = str(prefix)
        self.save_start = int(start)
        self.save_frequency = int(frequency)

    def save(self, filename: Path):
        """
        Save a model to disk.

        Args:
            filename: Path to store the model.
        """
        filename = to_path(filename)
        create_directory(filename)
        torch.save(self.state_dict(), str(filename))

        # save descriptor mean and stdev
        fname = filename.parent.joinpath("mean_and_stdev.pkl")
        self.descriptor.dump_mean_stdev(fname)

    def load(self, filename: Path, mode: str = "train"):
        """
        Load a model on disk into memory.

        Args:
            filename: Path where the model is stored.
            mode: Purpose of the loaded model. Should be either `train` or `eval`.
        """
        filename = to_path(filename)
        self.load_state_dict(torch.load(str(filename)))

        if mode == "train":
            self.train()
        elif mode == "eval":
            self.eval()
        else:
            raise ModelTorchError(
                'Unrecognized mode "{}" in model.load().'.format(mode)
            )

        # load descriptor mean and stdev
        fname = filename.parent.joinpath("mean_and_stdev.pkl")
        self.descriptor.load_mean_stdev(fname)

    def write_kim_model(self, path: Path = None):
        raise NotImplementedError("`write_kim_model` not implemented.")


class ModelTorchError(Exception):
    def __init__(self, msg):
        super(ModelTorchError, self).__init__(msg)
        self.msg = msg



import logging
import sys
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, TextIO, Tuple, Union

import numpy as np
from kliff.dataset.dataset import Configuration
from kliff.models.parameter import OptimizingParameters, Parameter
from kliff.utils import yaml_dump, yaml_load

logger = logging.getLogger(__name__)


class ComputeArguments:
    """
    Compute property (e.g. energy, forces, and stress) for a configuration.

    This is the base class for other compute arguments. Typically, a user will not
    directly use this.

    Args:
        conf: atomic configurations
        supported_species: species supported by the potential model, with chemical
            symbol as key and integer code as value.
        influence_distance: influence distance (aka cutoff distance) to calculate neighbors
        compute_energy: whether to compute energy
        compute_forces: whether to compute forces
        compute_stress: whether to compute stress
    """

    implemented_property = []

    def __init__(
        self,
        conf: Configuration,
        supported_species: Dict[str, int],
        influence_distance: float,
        compute_energy: bool = True,
        compute_forces: bool = True,
        compute_stress: bool = False,
    ):

        self.conf = conf
        self.supported_species = supported_species
        self.influence_distance = influence_distance
        self.compute_energy = compute_energy
        self.compute_forces = compute_forces
        self.compute_stress = compute_stress

        self.compute_property = self._check_compute_property()
        self.results = {p: None for p in self.implemented_property}

    def compute(self, params: Dict[str, Parameter]):
        """
        Compute the properties required by the compute flags, and store them in
        self.results.

        Args:
            params: the parameters of the model.

        Example:
            energy = a_func_to_compute_energy()
            forces = a_func_to_compute_forces()
            stress = a_func_to_compute_stress()
            self.results['energy'] = energy
            self.results['forces'] = forces
            self.results['stress'] = stress
        """
        raise NotImplementedError('"compute" method not implemented.')

    def get_compute_flag(self, name: str) -> bool:
        """
        Check whether the model is asked to compute property.

        Args:
            name: name of the property, e.g. energy, forces, and stresses
        """
        if name in self.compute_property:
            return True
        else:
            return False

    def get_property(self, name: str) -> Any:
        """
        Get a property by name.

        Args:
            name: name of the property, e.g. energy, forces, and stresses
        """
        if name not in self.compute_property:
            ModelError(f"Model not initialized to comptue `{name}`.")
        return self.results[name]

    def get_energy(self) -> float:
        """
        Potential energy.
        """
        return self.get_property("energy")

    def get_forces(self) -> np.ndarray:
        """
        2D array of shape (N,3) of the forces on atoms, where N is the number of atoms
        in the configuration.
        """
        return self.get_property("forces")

    def get_stress(self) -> np.ndarray:
        """
        1D array of the virial stress, in Voigt notation.
        """
        return self.get_property("stress")

    def get_prediction(self) -> np.ndarray:
        """
        1D array of prediction from the model for the configuration.
        """
        if self.compute_energy:
            energy = self.results["energy"]
            pred = np.asarray([energy])
        else:
            pred = np.asarray([])

        if self.compute_forces:
            forces = self.results["forces"]
            pred = np.concatenate((pred, forces.ravel()))

        if self.compute_stress:
            stress = self.results["stress"]
            pred = np.concatenate((pred, stress))

        return pred

    def get_reference(self) -> np.ndarray:
        """
        1D array of reference values for the configuration.
        """
        if self.compute_energy:
            energy = self.conf.energy
            ref = np.asarray([energy])
        else:
            ref = np.asarray([])

        if self.compute_forces:
            forces = self.conf.forces
            ref = np.concatenate((ref, forces.ravel()))

        if self.compute_stress:
            stress = self.conf.stress
            ref = np.concatenate((ref, stress))

        return ref

    def _check_compute_property(self):
        compute_property = []
        if self.compute_energy:
            if "energy" not in self.implemented_property:
                raise NotImplementedError("`Energy` not implemented in model.")
            else:
                compute_property.append("energy")
        if self.compute_forces:
            if "forces" not in self.implemented_property:
                raise NotImplementedError("`Forces` not implemented in model.")
            else:
                compute_property.append("forces")
        if self.compute_stress:
            if "stress" not in self.implemented_property:
                raise NotImplementedError("`Stress` not implemented in model.")
            else:
                compute_property.append("stress")

        return compute_property


class Model:
    """
    Base class for all physics-motivated models.

    Typically, a user will not directly use this.

    Args:
        model_name: name of the model.

        params_relation_callback: A callback function to set the relations between
            parameters, which are called each minimization step after the optimizer
            updates the parameters. The function with be given a dictionary of
            :meth:`~kliff.model.parameter.Parameter` as argument, which can
            then be manipulated to set relations between parameters.

            Example:
                In the following example, we set the value of ``B[0]`` to ``2 * A[0]``::

                    def params_relation(model_params):
                        A = model_params['A']
                        B = model_params['B']
                        B[0] = 2*A[0]
    """

    def __init__(
        self,
        model_name: str = None,
        params_relation_callback: Optional[Callable] = None,
    ):
        self.model_name = model_name
        self.params_relation_callback = params_relation_callback

        self.model_params = self.init_model_params()
        self.opt_params = OptimizingParameters(self.model_params)
        self.influence_distance = self.init_influence_distance()
        self.supported_species = self.init_supported_species()

    def init_model_params(self, *args, **kwargs) -> Dict[str, Parameter]:
        """
        Initialize the parameters of the model.

        Should return a dictionary of parameters.

        Example:
            model_params = {"sigma": Parameter([0.5]
                      "epsilon": Parameter([0.4])

            return model_params
        """
        raise NotImplementedError("`init_model_params` not implemented.")

    def init_influence_distance(self, *args, **kwargs) -> float:
        """
        Initialize the influence distance (aka cutoff distance) of the model.

        This is used to compute the neighbor list of each atom.

        Example:
            return 5.0
        """
        raise NotImplementedError("`init_influence_distance` not implemented.")

    def init_supported_species(self, *args, **kwargs) -> Dict[str, int]:
        """
        Initialize the supported species of the model.

        Should return a dict with chemical symbol as key and integer code as value.

        Example:
            return {"C:0, "O":1}
        """
        raise NotImplementedError("`init_supported_species` not implemented.")

    def get_compute_argument_class(self):
        # NOTE, change to the compute argument class as needed
        # return ComputeArguments
        raise NotImplementedError("`get_compute_argument_class` not implemented.")

    def write_kim_model(self, path: Path = None):
        raise NotImplementedError("`write_kim_model` not implemented yet.")

    def get_influence_distance(self) -> float:
        """
        Return influence distance (aka cutoff distance) of the model.
        """
        return self.influence_distance

    def get_supported_species(self) -> Dict[str, int]:
        """
        Return supported species of the model, a dict with chemical symbol as key and
        integer code as value.
        """
        return self.supported_species

    def get_model_params(self) -> Dict[str, Parameter]:
        """
        Return all parameters of the model.
        """
        return self.model_params

    def echo_model_params(
        self, filename: Union[Path, TextIO, None] = sys.stdout
    ) -> str:
        """
        Echo the model parameters.

        Args:
            filename: Path to write the model parameter info (e.g. sys.stdout) if `None`,
                do not write.

        Returns:
            model parameters info in a string
        """
        s = "#" + "=" * 80 + "\n"
        s += "# Available parameters to optimize.\n"

        name = self.__class__.__name__ if self.model_name is None else self.model_name
        s += f"# Model: {name}\n"

        s += "#" + "=" * 80 + "\n\n"

        for name, p in self.model_params.items():
            s += f"name: {name}\n"
            s += f"value: {p.value}\n"
            s += f"size: {len(p)}\n\n"

        if filename is not None:
            if isinstance(filename, (str, Path)):
                with open(filename, "w") as f:
                    f.write(s)
            else:
                filename.write(s)

        return s

    def read_opt_params(self, filename: Path):
        """
        Read optimizing parameters from a file.

        Each parameter is a 1D array, and each component of the parameter array should be
        listed in a new line. Each line can contains 1, 2, or 3 elements, described in
        details below:

        1st element: float or `DEFAULT`
            Initial guess of the parameter component. If `DEFAULT` (case insensitive), the
            value from the calculator is used as the initial guess.

        The 2nd and 3rd elements are optional.

        If 2 elements are provided:

        2nd element: `FIX` (case insensitive)
            If `FIX`, the corresponding component will not be optimized.

        If 3 elements are provided:

        2nd element: float or `INF` (case insensitive)
            Lower bound of the parameter. `INF` indicates that the lower bound is
            negative infinite, i.e. no lower bound is applied.

        3rd element: float or `INF` (case insensitive)
            Upper bound of the parameter. `INF` indicates that the upper bound is
            positive infinite, i.e. no upper bound is applied.

        Instead of reading fitting parameters from a file, you can also setting them
        using a dictionary by calling the `set_opt_params()` or `set_one_opt_params()`
        method.

        Args:
            filename: path to file that includes the fitting parameters.

        Example:

            # put the below in a file, say `model_params.txt` and you can read the fitting
            # parameters by this_class.read(filename="model_params.txt")

            A
            DEFAULT
            1.1

            B
            DEFAULT FIX
            1.1     FIX

            C
            DEFAULT  0.1  INF
            1.0      INF  2.1
            2.0      FIX
        """
        return self.opt_params.read(filename)

    def set_opt_params(self, **kwargs):
        """
        Set the parameters that will be optimized.

        One or more parameters can be set. Each argument is for one parameter, where the
        argument name is the parameter name, the value of the argument is the
        settings(including initial value, fix flag, lower bound, and upper bound).

        The value of the argument should be a list of list, where each inner list is for
        one component of the parameter, which can contain 1, 2, or 3 elements.
         See `~kliff.model.model.Model.read_opt_params()` for the options of the elements.

        Example:
           instance.set(A=[['DEFAULT'], [2.0, 1.0, 3.0]], B=[[1.0, 'FIX'], [2.0, 'INF', 3.0]])
        """
        self.opt_params.set(**kwargs)

        # reset influence distance in case it depends on parameters and changes
        self.init_influence_distance()

    def set_one_opt_param(self, name: str, settings: List[List[Any]]):
        """
        Set one parameter that will be optimized.

        The name of the parameter should be given as the first entry of a list (or tuple),
        and then each data line should be given in in a list.

        Args:
            name: name of a fitting parameter
            settings: initial value, flag to fix a parameter, lower and upper bounds of a
                parameter.

        Example:
            name = 'param_A'
            settings = [['default', 0, 20], [2.0, 'fix'], [2.2, 'inf', 3.3]]
            instance.set_one(name, settings)
        """
        self.opt_params.set_one(name, settings)

        # reset influence distance in case it depends on parameters and changes
        self.init_influence_distance()

    def echo_opt_params(self, filename: [Path, TextIO, None] = sys.stdout):
        """
        Echo the optimizing parameter to a file.
        """
        return self.opt_params.echo_opt_params(filename)

    def get_num_opt_params(self) -> int:
        """
        Number of optimizing parameters.

        This is the total number of model parameter components. For example,
        if the model has two parameters set to be optimized and each have two components,
        this will be four.
        """
        return self.opt_params.get_num_opt_params()

    def get_opt_params(self) -> np.ndarray:
        """
         Nest all optimizing parameter values (except the fixed ones) to a 1D array.

        The obtained values can be provided to the optimizer as the starting parameters.

        This is the opposite operation of update_model_params().

        Returns:
            opt_params: A 1D array of nested optimizing parameter values.
        """
        return self.opt_params.get_opt_params()

    def update_model_params(self, params: Sequence[float]):
        """
        Update the optimizing parameter values from a sequence of float.

        This is the opposite operation of get_opt_params().

        Note, self.model_params will be updated as well, since self.opt_params is
        initialized from self.model_params without copying the `Parameter` instance.

        Args:
            params: updated parameter values from the optimizer.
        """
        self.opt_params.update_opt_params(params)

        if self.params_relation_callback is not None:
            self.params_relation_callback(self.model_params)

    def get_opt_param_name_value_and_indices(
        self, index: int
    ) -> Tuple[str, float, int, int]:
        """
        Get the `name`, `value`, `parameter_index`, and `component_index` of optimizing
        parameter in slot `index`.

        Args:
            index: slot index of the optimizing parameter
        """
        return self.opt_params.get_opt_param_name_value_and_indices(index)

    def get_opt_params_bounds(self) -> List[Tuple[int, int]]:
        """
        Get the lower and upper bounds of optimizing parameters.
        """
        return self.opt_params.get_opt_params_bounds()

    def has_opt_params_bounds(self) -> bool:
        """
        Whether bounds are set for some parameters.
        """
        return self.opt_params.has_opt_params_bounds()

    def save(self, filename: Path = "trained_model.yaml"):
        """
        Save a model to disk.

        Args:
            filename: Path where to store the model.
        """
        d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "opt_params": self.opt_params.as_dict(),
        }
        yaml_dump(d, filename)

    def load(self, filename: Path = "trained_model.yaml"):
        """
        Load a model on disk into memory.

        Args:
            filename: Path where the model is stored.
        """
        d = yaml_load(filename)
        self.opt_params = OptimizingParameters.from_dict(d["opt_params"])

        # Set model_params to opt_params.model_params since they should share the
        # underlying `Parameter` objects
        self.model_params = self.opt_params.model_params


class ModelError(Exception):
    def __init__(self, msg):
        super(ModelError, self).__init__(msg)
        self.msg = msg



from ..utils import torch_available
from .calculator import Calculator

__all__ = ["Calculator"]


if torch_available():
    from .calculator_torch import CalculatorTorch, CalculatorTorchDDPCPU

    __all__.extend(["CalculatorTorch", "CalculatorTorchDDPCPU"])



import logging

import numpy as np

from ...neighbor import NeighborList
from ..descriptor import Descriptor, generate_full_cutoff, generate_species_code
from . import bs

logger = logging.getLogger(__name__)


class Bispectrum(Descriptor):
    """Bispectrum descriptor.

    Process dataset to generate fingerprints using the Bispectrum descriptor as discussed
    in [Bartok2010]_ and [Thompson2015]_.

    Parameters
    ----------
    cut_dists: dict
        Cutoff distances, with key of the form ``A-B`` where ``A`` and ``B`` are atomic
        species string, and value should be a float.

    cut_name: str
        Name of the cutoff function.

    hyperparams: dict
        A dictionary of the hyperparams of the descriptor.

    normalize: bool (optional)
        If ``True``, the fingerprints is centered and normalized according to:
        ``zeta = (zeta - mean(zeta)) / stdev(zeta)``

    dtype: np.dtype
        Data type for the generated fingerprints, such as ``np.float32`` and
        ``np.float64``.

    Example
    -------
    >>> cut_name = 'cos'
    >>> cut_dists = {'C-C': 5.0, 'C-H': 4.5, 'H-H': 4.0}
    >>> hyperparams = {'jmax': 4, 'weight': {'C':1.0, 'H':1.0}}
    >>> desc = Bispectrum(cut_dists, cut_name, hyperparams)

    References
    ----------
    .. [Bartok2010] Bartók, Albert P., Mike C. Payne, Risi Kondor, and Gábor Csányi.
       "Gaussian approximation potentials: The accuracy of quantum mechanics, without
       the electrons." Physical review letters 104, no. 13 (2010): 136403.
    .. [Thompson2015] Thompson, Aidan P., Laura P. Swiler, Christian R. Trott,
       Stephen M. Foiles, and Garritt J. Tucker. "Spectral neighbor analysis method
       for automated generation of quantum-accurate interatomic potentials." Journal
       of Computational Physics 285 (2015): 316-330.
    """

    def __init__(
        self,
        cut_dists,
        cut_name=None,
        hyperparams=None,
        normalize=True,
        dtype=np.float32,
    ):
        super(Bispectrum, self).__init__(
            cut_dists, cut_name, hyperparams, normalize, dtype
        )

        self.update_hyperparams(self.hyperparams)

        # init cdesc
        rfac0 = self.hyperparams["rfac0"]
        jmax = self.hyperparams["jmax"]
        diagonalstyle = self.hyperparams["diagonalstyle"]
        rmin0 = self.hyperparams["rmin0"]
        switch_flag = self.hyperparams["switch_flag"]
        bzero_flag = self.hyperparams["bzero_flag"]
        use_shared_arrays = 0

        self._cdesc = bs.Bispectrum(
            rfac0,
            2 * jmax,
            diagonalstyle,
            use_shared_arrays,
            rmin0,
            switch_flag,
            bzero_flag,
        )

        self._set_cutoff()

        self._set_hyperparams()

    def transform(self, conf, grad=False):

        # neighbor list
        infl_dist = max(self.cutoff.values())

        nei = NeighborList(conf, infl_dist, padding_need_neigh=False)

        coords = nei.coords
        image = nei.image
        species = np.asarray([self.species_code[i] for i in nei.species], dtype=np.intc)

        numneigh, neighlist = nei.get_numneigh_and_neighlist_1D()

        Natoms = len(coords)
        Ncontrib = conf.get_num_atoms()
        Ndesc = self.get_size()

        if grad:
            zeta, dzeta_dr = self._cdesc.compute_zeta_and_dzeta_dr(
                coords, species, neighlist, numneigh, image, Natoms, Ncontrib, Ndesc
            )
            # reshape to 4D array
            dzeta_dr = dzeta_dr.reshape(Ncontrib, Ndesc, Ncontrib, 3)
        else:
            zeta = self._cdesc.compute_zeta(
                coords, species, neighlist, numneigh, image, Natoms, Ncontrib, Ndesc
            )
            dzeta_dr = None

        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug(
                "\n" + "=" * 25 + "descriptor values (no normalization)" + "=" * 25
            )

            logger.debug("\nconfiguration name: %s", conf.get_identifier())

            logger.debug("\natom id    descriptor values ...")

            for i, line in enumerate(zeta):
                s = "\n{}    ".format(i)
                for j in line:
                    s += "{:.15g} ".format(j)
                logger.debug(s)

        return zeta, dzeta_dr

    def update_hyperparams(self, params):
        """Update the hyperparameters based on the input at initialization."""
        default_hyperparams = {
            "jmax": 4,
            "rfac0": 0.99363,
            "diagonalstyle": 3,
            "rmin0": 0,
            "switch_flag": 1,
            "bzero_flag": 0,
            "weight": None,
        }

        if params is not None:
            for key, value in params.items():
                if key not in default_hyperparams:
                    raise BispectrumError(
                        'Hyperparameter "{}" not supported by this descriptor.'.format(
                            key
                        )
                    )
                else:
                    default_hyperparams[key] = value

        self.hyperparams = default_hyperparams

    def _set_cutoff(self):
        supported = ["cos"]

        if self.cut_name is None:
            self.cut_name = supported[0]

        if self.cut_name not in supported:
            spd = ['"{}", '.format(s) for s in supported]
            raise BispectrumError(
                'Cutoff "{}" not supported by this descriptor. Use {}.'.format(
                    self.cut_name, spd
                )
            )

        self.cutoff = generate_full_cutoff(self.cut_dists)

        self.species_code = generate_species_code(self.cut_dists)

        num_species = len(self.species_code)

        rcutsym = np.zeros([num_species, num_species], dtype=np.double)

        for si, i in self.species_code.items():
            for sj, j in self.species_code.items():
                rcutsym[i][j] = self.cutoff[si + "-" + sj]

        self._cdesc.set_cutoff(self.cut_name, rcutsym)

    def _set_hyperparams(self):
        weight_in = self.hyperparams["weight"]

        if weight_in is None:
            weight = np.ones(len(self.species_code), dtype=np.double)
        else:
            weight = np.zeros(len(self.species_code), dtype=np.double)

            for spec, code in self.species_code.items():
                try:
                    weight[code] = weight_in[spec]
                except KeyError:
                    raise BispectrumError(
                        '"weight" for species "{}" not provided.'.format(spec)
                    )

        self._cdesc.set_weight(weight)

    def get_size(self):
        """
        Return the size of descriptor.
        """
        diagonal = self.hyperparams["diagonalstyle"]
        twojmax = int(2 * self.hyperparams["jmax"])

        N = 0
        for j1 in range(0, twojmax + 1):
            if diagonal == 2:
                N += 1
            elif diagonal == 1:
                for j in range(0, min(twojmax, 2 * j1) + 1, 2):
                    N += 1
            elif diagonal == 0:
                for j2 in range(0, j1 + 1):
                    for j in range(j1 - j2, min(twojmax, j1 + j2) + 1, 2):
                        N += 1
            elif diagonal == 3:
                for j2 in range(0, j1 + 1):
                    for j in range(j1 - j2, min(twojmax, j1 + j2) + 1, 2):
                        if j >= j1:
                            N += 1
        return N


class BispectrumError(Exception):
    def __init__(self, msg):
        super(BispectrumError, self).__init__(msg)

        self.msg = msg



"""
.. _tut_nn:

Train a neural network potential
================================

In this tutorial, we train a neural network (NN) potential for silicon.

"""


##########################################################################################
# We are going to fit the NN potential to a training set of energies and forces from
# compressed and stretched diamond silicon structures (the same training set used in
# :ref:`tut_kim_sw`).
# Download the training set :download:`Si_training_set.tar.gz <https://raw.githubusercontent.com/mjwen/kliff/master/examples/Si_training_set.tar.gz>`
# (It will be automatically downloaded if it is not present.)
# The data is stored in **extended xyz** format, and see :ref:`doc.dataset` for more
# information of this format.
#
# .. warning::
#     The ``Si_training_set`` is just a toy data set for the purpose to demonstrate how to
#     use KLIFF to train potentials. It should not be used to train any potential for real
#     simulations.
#
# Let's first import the modules that will be used in this example.

from kliff import nn
from kliff.calculators import CalculatorTorch
from kliff.dataset import Dataset
from kliff.descriptors import SymmetryFunction
from kliff.loss import Loss
from kliff.models import NeuralNetwork
from kliff.utils import download_dataset

##########################################################################################
# Model
# -----
#
# For a NN model, we need to specify the descriptor that transforms atomic environment
# information to the fingerprints, which the NN model uses as the input. Here, we use the
# symmetry functions proposed by Behler and coworkers.

descriptor = SymmetryFunction(
    cut_name="cos", cut_dists={"Si-Si": 5.0}, hyperparams="set51", normalize=True
)


##########################################################################################
# The ``cut_name`` and ``cut_dists`` tell the descriptor what type of cutoff function to
# use and what the cutoff distances are. ``hyperparams`` specifies the set of
# hyperparameters used in the symmetry function descriptor. If you prefer, you can provide
# a dictionary of your own hyperparameters. And finally, ``normalize`` informs that the
# generated fingerprints should be normalized by first subtracting the mean and then
# dividing the standard deviation. This normalization typically makes it easier to
# optimize NN model.
#
# We can then build the NN model on top of the descriptor.

N1 = 10
N2 = 10
model = NeuralNetwork(descriptor)
model.add_layers(
    # first hidden layer
    nn.Linear(descriptor.get_size(), N1),
    nn.Tanh(),
    # second hidden layer
    nn.Linear(N1, N2),
    nn.Tanh(),
    # output layer
    nn.Linear(N2, 1),
)
model.set_save_metadata(prefix="./kliff_saved_model", start=5, frequency=2)


##########################################################################################
# In the above code, we build a NN model with an input layer, two hidden layer, and an
# output layer. The ``descriptor`` carries the information of the input layer, so it is
# not needed to be specified explicitly. For each hidden layer, we first do a linear
# transformation using ``nn.Linear(size_in, size_out)`` (essentially carrying out :math:`y
# = xW+b`, where :math:`W` is the weight matrix of size ``size_in`` by ``size_out``, and
# :math:`b` is a vector of size ``size_out``. Then we apply the hyperbolic tangent
# activation function ``nn.Tanh()`` to the output of the Linear layer (i.e. :math:`y`) so
# as to add the nonlinearity. We use a Linear layer for the output layer as well, but
# unlike the hidden layer, no activation function is applied here. The input size
# ``size_in`` of the first hidden layer must be the size of the descriptor, which is
# obtained using ``descriptor.get_size()``. For all other layers (hidden or output), the
# input size must be equal to the output size of the previous layer. The ``out_size`` of
# the output layer must be 1 such that the output of the NN model gives the energy of the
# atom.
#
# The ``set_save_metadata`` function call informs where to save intermediate models during
# the optimization (discussed below), and what the starting epoch and how often to save
# the model.
#
#
# Training set and calculator
# ---------------------------
#
# The training set and the calculator are the same as explained in :ref:`tut_kim_sw`. The
# only difference is that we need to use the
# :mod:`~kliff.calculators.CalculatorTorch()`, which is targeted for the NN model.
# Also, its ``create()`` method takes an argument ``reuse`` to inform whether to reuse the
# fingerprints generated from the descriptor if it is present.

# training set
dataset_path = download_dataset(dataset_name="Si_training_set")
dataset_path = dataset_path.joinpath("varying_alat")
tset = Dataset(dataset_path)
configs = tset.get_configs()

# calculator
calc = CalculatorTorch(model)
calc.create(configs, reuse=False)


##########################################################################################
# Loss function
# -------------
#
# KLIFF uses a loss function to quantify the difference between the training data and
# potential predictions and uses minimization algorithms to reduce the loss as much as
# possible. In the following code snippet, we create a loss function that uses the
# ``Adam`` optimizer to minimize it. The Adam optimizer supports minimization using
# `mini-batches` of data, and here we use ``100`` configurations in each minimization step
# (the training set has a total of 400 configurations as can be seen above), and run
# through the training set for ``10`` epochs. The learning rate ``lr`` used here is
# ``0.001``, and typically, one may need to play with this to find an acceptable one that
# drives the loss down in a reasonable time.

loss = Loss(calc, residual_data={"forces_weight": 0.3})
result = loss.minimize(method="Adam", num_epochs=10, batch_size=100, lr=0.001)


##########################################################################################
# We can save the trained model to disk, and later can load it back if we want. We can
# also write the trained model to a KIM model such that it can be used in other simulation
# codes such as LAMMPS via the KIM API.

model.save("final_model.pkl")
loss.save_optimizer_state("optimizer_stat.pkl")

model.write_kim_model()


##########################################################################################
# .. note::
#    Now we have trained an NN for a single specie Si. If you have multiple species in
#    your system and want to use different parameters for different species,
#    take a look at the :ref:`tut_nn_multi_spec` example.
#



"""
.. _tut_kim_sw:

Train a Stillinger-Weber potential
==================================

In this tutorial, we train a Stillinger-Weber (SW) potential for silicon that is archived
on OpenKIM_.
"""


##########################################################################################
# Before getting started to train the SW model, let's first install the SW model::
#
#    $ kim-api-collections-management install user SW_StillingerWeber_1985_Si__MO_405512056662_005
#
# .. seealso::
#    This installs the model and its driver into the ``User Collection``. See
#    :ref:`install_model` for more information about installing KIM models.
#
# We are going to create potentials for diamond silicon, and fit the potentials to a
# training set of energies and forces consisting of compressed and stretched diamond
# silicon structures, as well as configurations drawn from molecular dynamics trajectories
# at different temperatures.
# Download the training set :download:`Si_training_set.tar.gz
# <https://raw.githubusercontent.com/mjwen/kliff/master/examples/Si_training_set.tar.gz>`.
# (It will be automatically downloaded if not present.)
# The data is stored in # **extended xyz** format, and see :ref:`doc.dataset` for more
# information of this format.
#
# .. warning::
#    The ``Si_training_set`` is just a toy data set for the purpose to demonstrate how to
#    use KLIFF to train potentials. It should not be used to train any potential for real
#    simulations.
#
# Let's first import the modules that will be used in this example.

from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.loss import Loss
from kliff.models import KIMModel
from kliff.utils import download_dataset

##########################################################################################
# Model
# -----
#
# We first create a KIM model for the SW potential, and print out all the available
# parameters that can be optimized (we call this ``model parameters``).

model = KIMModel(model_name="SW_StillingerWeber_1985_Si__MO_405512056662_005")
model.echo_model_params()


##########################################################################################
# The output is generated by the last line, and it tells us the ``name``, ``value``,
# ``size``, ``data type`` and a ``description`` of each parameter.
#
# .. note::
#    You can provide a ``path`` argument to the method ``echo_model_params(path)`` to
#    write the available parameters information to a file indicated by ``path``.
#
# .. note::
#    The available parameters information can also by obtained using the **kliff**
#    :ref:`cmdlntool`:
#    ``$ kliff model --echo-params SW_StillingerWeber_1985_Si__MO_405512056662_005``
#
# Now that we know what parameters are available for fitting, we can optimize all or a
# subset of them to reproduce the training set.

model.set_opt_params(
    A=[[5.0, 1.0, 20]], B=[["default"]], sigma=[[2.0951, "fix"]], gamma=[[1.5]]
)
model.echo_opt_params()


##########################################################################################
# Here, we tell KLIFF to fit four parameters ``B``, ``gamma``, ``sigma``, and ``A`` of the
# SW model. The information for each fitting parameter should be provided as a list of
# list, where the size of the outer list should be equal to the ``size`` of the parameter
# given by ``model.echo_model_params()``. For each inner list, you can provide either one,
# two, or three items.
#
# - One item. You can use a numerical value (e.g. ``gamma``) to provide an initial guess
#   of the parameter. Alternatively, the string ``'default'`` can be provided to use the
#   default value in the model (e.g. ``B``).
#
# - Two items. The first item should be a numerical value and the second item should be
#   the string ``'fix'`` (e.g. ``sigma``), which tells KLIFF to use the value for the
#   parameter, but do not optimize it.
#
# - Three items. The first item can be a numerical value or the string ``'default'``,
#   having the same meanings as the one item case. In the second and third items, you can
#   list the lower and upper bounds for the parameters, respectively. A bound could be
#   provided as a numerical values or ``None``. The latter indicates no bound is applied.
#
# The call of ``model.echo_opt_params()`` prints out the fitting parameters that we
# require KLIFF to optimize. The number ``1`` after the name of each parameter indicates
# the size of the parameter.
#
# .. note::
#    The parameters that are not included as a fitting parameter are fixed to the default
#    values in the model during the optimization.
#
#
# Training set
# ------------
#
# KLIFF has a :class:`~kliff.dataset.Dataset` to deal with the training data (and possibly
# test data). For the silicon training set, we can read and process the files by:

dataset_path = download_dataset(dataset_name="Si_training_set")
tset = Dataset(dataset_path)
configs = tset.get_configs()


##########################################################################################
# The ``configs`` in the last line is a list of :class:`~kliff.dataset.Configuration`.
# Each configuration is an internal representation of a processed **extended xyz** file,
# hosting the species, coordinates, energy, forces, and other related information of a
# system of atoms.
#
#
# Calculator
# ----------
#
# :class:`~kliff.calculator.Calculator` is the central agent that exchanges information
# and orchestrate the operation of the fitting process. It calls the model to compute the
# energy and forces and provide this information to the `Loss function`_ (discussed below)
# to compute the loss. It also grabs the parameters from the optimizer and update the
# parameters stored in the model so that the up-to-date parameters are used the next time
# the model is evaluated to compute the energy and forces. The calculator can be created
# by:

calc = Calculator(model)
calc.create(configs)


##########################################################################################
# where ``calc.create(configs)`` does some initializations for each
# configuration in the training set, such as creating the neighbor list.
#
#
# Loss function
# -------------
#
# KLIFF uses a loss function to quantify the difference between the training set data and
# potential predictions and uses minimization algorithms to reduce the loss as much as
# possible. KLIFF provides a large number of minimization algorithms by interacting with
# SciPy_. For physics-motivated potentials, any algorithm listed on
# `scipy.optimize.minimize`_ and `scipy.optimize.least_squares`_ can be used. In the
# following code snippet, we create a loss of energy and forces, where the residual
# function uses an ``energy_weight`` of ``1.0`` and a ``forces_weight`` of ``0.1``, and
# ``2`` processors will be used to calculate the loss. The ``L-BFGS-B`` minimization
# algorithm is applied to minimize the loss, and the minimization is allowed to run for
# a max number of 100 iterations.

steps = 100
residual_data = {"energy_weight": 1.0, "forces_weight": 0.1}
loss = Loss(calc, residual_data=residual_data, nprocs=2)
loss.minimize(method="L-BFGS-B", options={"disp": True, "maxiter": steps})


##########################################################################################
# The minimization stops after running for 27 steps.  After the minimization, we'd better
# save the model, which can be loaded later for the purpose to do a retraining or
# evaluations. If satisfied with the fitted model, you can also write it as a KIM model
# that can be used with LAMMPS_, GULP_, ASE_, etc. via the kim-api_.

model.echo_opt_params()
model.save("kliff_model.yaml")
model.write_kim_model()
# model.load("kliff_model.yaml")


##########################################################################################
# The first line of the above code generates the output.  A comparison with the original
# parameters before carrying out the minimization shows that we recover the original
# parameters quite reasonably. The second line saves the fitted model to a file named
# ``kliff_model.pkl`` on the disk, and the third line writes out a KIM potential named
# ``SW_StillingerWeber_1985_Si__MO_405512056662_005_kliff_trained``.
#
# .. seealso::
#    For information about how to load a saved model, see :ref:`doc.modules`.
#
#
# .. _OpenKIM: https://openkim.org
# .. _SciPy: https://scipy.org
# .. _scipy.optimize.minimize: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
# .. _scipy.optimize.least_squares: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html
# .. _kim-api: https://openkim.org/kim-api/
# .. _LAMMPS: https://lammps.sandia.gov
# .. _GULP: http://gulp.curtin.edu.au/gulp/
# .. _ASE: https://wiki.fysik.dtu.dk/ase/



import numpy as np
from kliff.parallel import parmap1, parmap2


def func(x, y, z=1):
    return x + y + z


def test_main():
    X = range(3)
    Y = range(3)
    Xp2 = [x + 2 for x in X]
    XpYp1 = [x + y + 1 for x, y in zip(X, Y)]

    results = parmap1(func, X, 1, nprocs=2)
    assert np.array_equal(results, Xp2)

    results = parmap1(func, X, 1, 1, nprocs=2)
    assert np.array_equal(results, Xp2)

    results = parmap1(func, zip(X, Y), nprocs=2, tuple_X=True)
    assert np.array_equal(results, XpYp1)

    results = parmap1(func, zip(X, Y), 1, nprocs=2, tuple_X=True)
    assert np.array_equal(results, XpYp1)

    results = parmap2(func, X, 1, nprocs=2)
    assert np.array_equal(results, Xp2)

    results = parmap2(func, X, 1, 1, nprocs=2)
    assert np.array_equal(results, Xp2)

    results = parmap2(func, zip(X, Y), nprocs=2, tuple_X=True)
    assert np.array_equal(results, XpYp1)

    results = parmap2(func, zip(X, Y), 1, nprocs=2, tuple_X=True)
    assert np.array_equal(results, XpYp1)



"""Format all .py and CPP source files using `black` and `clang-format` respectively.

This assumes both `black` (https://github.com/python/black) and `clang-format`
(https://clang.llvm.org/docs/ClangFormat.html) are installed.

To run:
$ python format_sources.py
"""

import os
import subprocess


def get_files(path, extension=[".cpp", ".hpp", ".h"]):
    all_srcs = []
    path = os.path.abspath(path)
    for root, dirs, files in os.walk(path):
        for f in files:
            ext = os.path.splitext(f)[1]
            if ext in extension:
                all_srcs.append(os.path.join(root, f))
    return all_srcs


def format_py_code(path):
    path = os.path.abspath(path)
    print(
        'Formating .py files in directory "{}" and its subdirectories using '
        '"black"...'.format(path)
    )
    subprocess.call(
        ["black", "--quiet", "--line-length", "90", "--skip-string-normalization", path]
    )
    print("Formatting .py files done.")


def format_cpp_code(path):
    path = os.path.abspath(path)
    print(
        'Formating CPP files in directory "{}" and its subdirectories using '
        '"clang-format"...'.format(path)
    )
    files = get_files(path)
    for f in files:
        subprocess.call(["clang-format", "-style=file", "-i", f])
    print("Formatting CPP files done.")


if __name__ == "__main__":
    format_py_code("../")
    format_cpp_code("../")



from pathlib import Path
from typing import Callable, Optional

import torch
from kliff.descriptors.descriptor import load_fingerprints
from torch.utils.data import Dataset


class FingerprintsDataset(Dataset):
    """
    Atomic environment fingerprints dataset used by torch models.

    Args:
        filename: to the fingerprints file.
        transform: transform to be applied on a sample.
    """

    def __init__(self, filename: Path, transform: Optional[Callable] = None):
        self.fp = load_fingerprints(filename)
        self.transform = transform

    def __len__(self):
        return len(self.fp)

    def __getitem__(self, index):
        sample = self.fp[index]
        if self.transform:
            sample = self.transform(sample)
        return sample


def fingerprints_collate_fn(batch):
    """
    Convert a batch of samples into tensor.

    Unlike the default collate_fn(), which stack samples in the batch (requiring each
    sample having the same dimension), this function does not do the stack.

    Args:
        batch: A batch of samples.

    Returns:
        A list of tensor.
    """
    tensor_batch = []
    for i, sample in enumerate(batch):
        tensor_sample = {}
        for key, value in sample.items():
            if type(value).__module__ == "numpy":
                value = torch.from_numpy(value)
            tensor_sample[key] = value
        tensor_batch.append(tensor_sample)

    return tensor_batch



import logging
import sys
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, TextIO, Tuple, Union

import numpy as np
from monty.json import MSONable

logger = logging.getLogger(__name__)


class Parameter(MSONable):
    """
    Potential parameter.

    To follow the KIM model paradigm, a parameter is a list of floats that can be
    adjusted to fit the potential to some parameters.

    This class is mainly by physically-motivated potentials.

    Args:
        value: parameter value, which can have multiple components.
        fixed: whether parameter component should be fixed (i.e. not used for fitting).
            Should have the same length of `value`. Default to not fixed for all
            components.
        lower_bound: lower bound of the allowed value for the parameter. Default to
            `None`, i.e. no lower bound is applied.
        upper_bound: upper bound of the allowed value for the parameter. Default to
            `None`, i.e. no lower bound is applied.
        name: name of the parameter.
        index: integer index of the parameter.
    """

    def __init__(
        self,
        value: Union[Sequence[float]],
        fixed: Optional[Sequence[bool]] = None,
        lower_bound: Optional[Sequence[float]] = None,
        upper_bound: Optional[Sequence[float]] = None,
        name: Optional[str] = None,
        index: Optional[int] = None,
    ):

        self._value = _check_shape(value, "parameter value")
        self._fixed = (
            [False] * len(self._value)
            if fixed is None
            else _check_shape(fixed, "fixed")
        )
        self._lower_bound = (
            [None] * len(self._value)
            if lower_bound is None
            else _check_shape(lower_bound, "lower_bound")
        )
        self._upper_bound = (
            [None] * len(self._value)
            if upper_bound is None
            else _check_shape(upper_bound, "upper_bound")
        )
        self._name = name
        self._index = index

    @property
    def value(self) -> List[float]:
        """
        Parameter value.
        """
        return self._value

    def set_value(self, index: int, v: float):
        """
        Set the value of a component.
        """
        self._value[index] = float(v)

    @property
    def fixed(self) -> List[bool]:
        """
        Whether each parameter component is fixed or not (i.e. allow to fitting or not).
        """
        return self._fixed

    def set_fixed(self, index: int, v: bool):
        """
        Set the fixed status of a component of the parameter.

        Args:
            index: index of the component
            v: fix status
        """
        self._fixed[index] = v

    @property
    def lower_bound(self) -> List[float]:
        """
        Lower bound of parameter.
        """
        return self._lower_bound

    def set_lower_bound(self, index: int, v: float):
        """
        Set the lower bound of a component of the parameter.

        Args:
            index: index of the component
            v: lower bound value
        """
        self._lower_bound[index] = v

    @property
    def upper_bound(self) -> List[float]:
        """
        Upper bound of parameter.
        """
        return self._upper_bound

    def set_upper_bound(self, index: int, v: float):
        """
        Set the upper bound of a component of the parameter.

        Args:
            index: index of the component
            v: upper bound value
        """
        self._upper_bound[index] = v

    @property
    def name(self) -> str:
        """
        Name of the parameter.
        """
        return self._name

    @property
    def index(self) -> int:
        """
        Integer index of the parameter.
        """
        return self._index

    def __len__(self):
        return len(self.value)

    def __getitem__(self, i: int):
        return self._value[i]

    def __setitem__(self, i: int, v: float):
        self._value[i] = float(v)

    def as_dict(self):
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "value": self._value,
            "fixed": self._fixed,
            "lower_bound": self._lower_bound,
            "upper_bound": self._upper_bound,
            "name": self._name,
            "index": self._index,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            value=d["value"],
            fixed=d["fixed"],
            lower_bound=d["lower_bound"],
            upper_bound=d["upper_bound"],
            name=d["name"],
            index=d["index"],
        )


class OptimizingParameters(MSONable):
    """
    A collection of paramters that will be optimized.

    This can be all the parameters of a model or a subset of the parameters of a model.
    The behavior of individual component of a parameter can also be controlled. For
    example, keep the 2nd component of a parameters fixed, while optimizing the other
    components.

    It interacts with optimizer to provide initial guesses of parameter values; it also
    receives updated parameters from the optimizer and update model parameters.

    Args:
        model_params: {name, parameter} all the parameters of a model. The attributes
            of these parameters will be modified to reflect whether it is optimized.
    """

    def __init__(self, model_params: Dict[str, Parameter]):
        self.model_params = model_params

        # list of optimizing param names
        self._params = []

        # individual components of parameters that are optimized
        self._index = []

    def read(self, filename: Path):
        """
        Read the parameters that will be optimized from a file.

        Each parameter is a 1D array, and each component of the parameter array should be
        listed in a new line. Each line can contains 1, 2, or 3 elements, described in
        details below:

        1st element: float or `DEFAULT`
            Initial guess of the parameter component. If `DEFAULT` (case insensitive), the
            value from the calculator is used as the initial guess.

        The 2nd and 3rd elements are optional.

        If 2 elements are provided:

        2nd element: `FIX` (case insensitive)
            If `FIX`, the corresponding component will not be optimized.

        If 3 elements are provided:

        2nd element: float or `INF` (case insensitive)
            Lower bound of the parameter. `INF` indicates that the lower bound is
            negative infinite, i.e. no lower bound is applied.

        3rd element: float or `INF` (case insensitive)
            Upper bound of the parameter. `INF` indicates that the upper bound is
            positive infinite, i.e. no upper bound is applied.

        Instead of reading fitting parameters from a file, you can also setting them
        using a dictionary by calling the `set()` method.

        Args:
            filename: path to file that includes the fitting parameters.

        Example:

            # put the below in a file, say `model_params.txt` and you can read the fitting
            # parameters by this_class.read(filename="model_params.txt")

            A
            DEFAULT
            1.1

            B
            DEFAULT FIX
            1.1     FIX

            C
            DEFAULT  0.1  INF
            1.0      INF  2.1
            2.0      FIX
        """

        with open(filename, "r") as fin:
            lines = fin.readlines()
            lines = _remove_comments(lines)

        num_line = 0
        while num_line < len(lines):
            name = lines[num_line].strip()
            num_line += 1

            if name in self._params:
                warnings.warn(
                    f"file `{filename}`, line {num_line}. Parameter `{name}` already "
                    f"set. Reset it.",
                    category=Warning,
                )

            if name not in self.model_params:
                raise ParameterError(
                    f"file `{filename}`, line {num_line}. Parameter `{name}` not "
                    f"supported."
                )

            num_components = len(self.model_params[name])

            settings = []
            for j in range(num_components):
                settings.append(lines[num_line].split())
                num_line += 1

            self.set_one(name, settings)

    def set(self, **kwargs):
        """
        Set the parameters that will be optimized.

        One or more parameters can be set. Each argument is for one parameter, where the
        argument name is the parameter name, the value of the argument is the
        settings(including initial value, fix flag, lower bound, and upper bound).

        The value of the argument should be a list of list, where each inner list is for
        one component of the parameter, which can contain 1, 2, or 3 elements.  See
        `~kliff.model.parameter.OptimizingParameters.read()` for the options of the
        elements.

        Example:
            instance.set(A=[['DEFAULT'], [2.0, 1.0, 3.0]], B=[[1.0, 'FIX'], [2.0, 'INF', 3.0]])
        """
        for name, settings in kwargs.items():
            if name in self._params:
                msg = f"Parameter `{name}` already set. Reset it."
                warnings.warn(msg, category=Warning)

            if name not in self.model_params:
                raise ParameterError(f"Parameter `{name}` not supported.")

            self.set_one(name, settings)

    def set_one(self, name: str, settings: List[List[Any]]):
        """
        Set one parameter that will be optimized.

        The name of the parameter should be given as the first entry of a list (or tuple),
        and then each data line should be given in in a list.

        Args:
            name: name of a fitting parameter
            settings: initial value, flag to fix a parameter, lower and upper bounds of a
                parameter.

        Example:
            name = 'param_A'
            settings = [['default', 0, 20], [2.0, 'fix'], [2.2, 'inf', 3.3]]
            instance.set_one(name, settings)
        """
        size = len(self.model_params[name])
        if len(settings) != size:
            raise ParameterError(
                f"Expect {size} components for parameter `{name}`, got {len(settings)}."
            )

        for j, line in enumerate(settings):
            num_items = len(line)
            if num_items == 1:
                self._read_1_item(name, j, line)
            elif num_items == 2:
                self._read_2_item(name, j, line)
            elif num_items == 3:
                self._read_3_item(name, j, line)
            else:
                raise ParameterError(
                    f"More than 3 elements listed at data line {j+1} for "
                    f"parameter `{name}`."
                )

            # probably setting cutoff?
            if "cut" in name:
                # and not fix it?
                if not (num_items == 2 and line[0].lower() == "fix"):
                    warnings.warn(
                        f"Parameter `{name}` seems to be a cutoff distance. KLIFF does "
                        f"not support optimizing cutoff. Remove it from your "
                        f"optimizing parameter or make it a `fix` parameter if you "
                        f"want to use a different cutoff distance from the default "
                        f"value in the model. Ignore this if `{name}` is not a cutoff "
                        f"distance."
                    )

        self._set_index(name)

        if name not in self._params:
            self._params.append(name)

    def echo_opt_params(
        self, filename: Union[Path, TextIO, None] = sys.stdout, echo_size: bool = True
    ) -> str:
        """
        Get the optimizing parameters as a string and/or print to file (stdout).

        Args:
            filename: Path to the file to output the optimizing parameters. If `None`,
                print to stdout.
            echo_size: Whether to print the size of parameters. (Each parameter
                may have one or more components).

        Returns:
            Optimizing parameters as a string.
        """

        s = "#" + "=" * 80 + "\n"
        s += "# Model parameters that are optimized.\n"
        s += "#" + "=" * 80 + "\n\n"

        for name in self._params:
            p = self.model_params[name]

            if echo_size:
                s += f"{name} {len(p)}\n"
            else:
                s += f"{name}\n"

            for i in range(len(p)):
                s += f"{p[i]:24.16e} "

                if p.fixed[i]:
                    s += "fix "

                lb = p.lower_bound[i]
                ub = p.upper_bound[i]
                has_lb = lb is not None
                has_ub = ub is not None
                has_bounds = has_lb or has_ub
                if has_bounds:
                    if has_lb:
                        s += f"{lb:24.16e} "
                    else:
                        s += "None "
                    if has_ub:
                        s += f"{ub:24.16e} "
                    else:
                        s += "None "
                s += "\n"

            s += "\n"

        if filename is not None:
            if isinstance(filename, (str, Path)):
                with open(filename, "w") as f:
                    f.write(s)
            else:
                filename.write(s)

        return s

    def get_num_opt_params(self) -> int:
        """
        Number of optimizing parameters.

        This is the total number of model parameter components. For example,
        if the model has two parameters set to be optimized and each have two components,
        this will be four.
        """
        return len(self._index)

    def get_opt_params(self) -> np.ndarray:
        """
        Nest all optimizing parameter values (except the fixed ones) to a 1D array.

        The obtained values can be provided to the optimizer as the starting parameters.

        This is the opposite operation of update_model_params().

        Returns:
            opt_params: A 1D array of nested optimizing parameter values.
        """
        params = []
        for idx in self._index:
            params.append(self.model_params[idx.name][idx.c_idx])
        if len(params) == 0:
            raise ParameterError("No parameters specified to optimize.")

        return np.asarray(params)

    def update_opt_params(self, params: Sequence[float]):
        """
        Update the optimizing parameter values from a sequence of float.

        This is the opposite operation of get_opt_params().

        Args:
            params: updated parameter values from the optimizer.
        """
        for k, val in enumerate(params):
            name = self._index[k].name
            c_idx = self._index[k].c_idx
            self.model_params[name][c_idx] = val

    def get_opt_param_name_value_and_indices(
        self, index: int
    ) -> Tuple[str, float, int, int]:
        """
        Get the `name`, `value`, `parameter_index`, and `component_index` of optimizing
        parameter in slot `index`.

        Args:
            index: slot index of the optimizing parameter
        """
        name = self._index[index].name
        p_idx = self._index[index].p_idx
        c_idx = self._index[index].c_idx
        value = self.model_params[name][c_idx]

        return name, value, p_idx, c_idx

    def get_opt_params_bounds(self) -> List[Tuple[int, int]]:
        """
        Get the lower and upper bounds of optimizing parameters.
        """
        bounds = []
        for idx in self._index:
            name = idx.name
            c_idx = idx.c_idx
            lower = self.model_params[name].lower_bound[c_idx]
            upper = self.model_params[name].upper_bound[c_idx]
            bounds.append([lower, upper])

        return bounds

    def has_opt_params_bounds(self) -> bool:
        """
        Whether bounds are set for some parameters.
        """
        bounds = self.get_opt_params_bounds()
        for low, up in bounds:
            if low is not None or up is not None:
                return True
        return False

    def _read_1_item(self, name, j, line):
        self._read_1st_item(name, j, line[0])

    def _read_2_item(self, name, j, line):
        self._read_1st_item(name, j, line[0])

        if line[1].lower() == "fix":
            self.model_params[name].set_fixed(j, True)
        else:
            raise ParameterError(f"Data at line {j+1} of {name} corrupted.")

    def _read_3_item(self, name, j, line):
        self._read_1st_item(name, j, line[0])

        value = self.model_params[name][j]

        # lower bound
        if (line[1] is not None) and (
            not (isinstance(line[1], str) and line[1].lower() == "none")
        ):
            try:
                self.model_params[name].set_lower_bound(j, float(line[1]))
            except ValueError as e:
                raise ParameterError(f"{e}.\nData at line {j+1} of {name} corrupted.")

            if float(line[1]) > value:
                raise ParameterError(
                    f"Lower bound ({line[1]}) larger than value({value}) at line {j+1} "
                    f"of parameter `{name}`"
                )

        # upper bound
        if (line[2] is not None) and (
            not (isinstance(line[2], str) and line[2].lower() == "none")
        ):
            try:
                self.model_params[name].set_upper_bound(j, float(line[2]))
            except ValueError as e:
                raise ParameterError(f"{e}.\nData at line {j+1} of {name} corrupted.")

            if float(line[2]) < value:
                raise ParameterError(
                    f"Upper bound ({line[2]}) smaller than value({value}) at line {j+1} "
                    f"of parameter `{name}`"
                )

    def _read_1st_item(self, name, j, first):
        if isinstance(first, str) and first.lower() == "default":
            pass
        else:
            try:
                self.model_params[name][j] = float(first)
            except ValueError as e:
                raise ParameterError(
                    f"{e}.\nData at line {j+1} of parameter `{name}` corrupted."
                )

    def _set_index(self, name: str):
        """
        Include parameter component that will be optimized (i.e. `fixed` is False) in
        the optimizing parameter index list.

        Given a parameter and its values such as:

        PARAM_FREE_B
        1.1
        2.2  fix
        4.4  3.3  5.5

        the first slot (1.1) and the third slot (4.4) will be included in self._index,
        and later be optimized.

        Args:
            name: name of the parameter
        """

        size = len(self.model_params[name])
        fix = self.model_params[name].fixed
        p_idx = self.model_params[name].index

        for c_idx in range(size):

            if not fix[c_idx]:
                idx = _Index(name, p_idx, c_idx)

                # check whether already in self._index
                already_in = False
                for k, i in enumerate(self._index):
                    if idx == i:
                        already_in = k
                        break
                if already_in is not False:
                    warnings.warn(
                        f"Parameter `{name}` component `{c_idx}` already set. Reset it.",
                        category=Warning,
                    )
                    self._index[already_in] = idx
                else:
                    self._index.append(idx)

    def as_dict(self):
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "model_params": {k: v.as_dict() for k, v in self.model_params.items()},
            "params": self._params,
        }

    @classmethod
    def from_dict(cls, d):
        c = cls(
            model_params={
                k: Parameter.from_dict(v) for k, v in d["model_params"].items()
            }
        )
        c._params = d["params"]
        for name in d["params"]:
            c._set_index(name)

        return c


class _Index:
    """
    Mapping of a component of the optimizing parameter list to the model parameters dict.
    """

    def __init__(self, name, parameter_index=None, component_index=None):
        self.name = name
        self.parameter_index = self.p_idx = parameter_index
        self.component_index = self.c_idx = component_index

    def set_parameter_index(self, index):
        self.parameter_index = self.p_idx = index

    def set_component_index(self, index):
        self.component_index = self.c_idx = index

    def __expr__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ != other.__dict__
        else:
            return True


def _remove_comments(lines: List[str]):
    """
    Remove lines in a string list that start with # and content after #.
    """
    processed_lines = []
    for line in lines:
        line = line.strip()
        if not line or line[0] == "#":
            continue
        if "#" in line:
            line = line[0 : line.index("#")]
        processed_lines.append(line)
    return processed_lines


def _check_shape(x: Any, key="parameter"):
    """Check x to be a 1D array or list-like sequence."""
    if isinstance(x, np.ndarray):
        x = x.tolist()
    if isinstance(x, Sequence):
        if any(isinstance(i, Sequence) for i in x):
            raise ParameterError(f"{key} should be a 1D array (or list).")
    else:
        raise ParameterError(f"{key} should be a 1D array (or list).")

    return x.copy()


class ParameterError(Exception):
    def __init__(self, msg):
        super(ParameterError, self).__init__(msg)
        self.msg = msg



import multiprocessing as mp
import random

import numpy as np


def parmap1(f, X, *args, tuple_X=False, nprocs=mp.cpu_count()):
    """
    Parallelism over data.

    This function mimics ``multiprocessing.Pool.map`` to allow extra arguments to be used
    for the function ``f``.

    Parameters
    ----------
    f: function
        The function that operates on the data.

    X: list
        Data to be parallelized.

    args: args
        Extra positional arguments needed by the function ``f``.

    tuple_X: bool
        This depends on ``X``. It should be set to ``True`` if multiple arguments are
        parallelized and set to ``False`` if only one argument is parallelized. See
        ``Example`` below.

    nprocs: int
        Number of processors to use.

    Return
    ------
    list
        A list of results, corresponding to ``X``.

    Note
    ----
    The data is put into a job queue, a worker process gets a piece of the data to work
    on, the worker pushes the result back to the manager through another queue, and then
    get another piece of data until the job queue is empty.  So, in principle, there will
    not be idle worker and it should be faster than :meth:`kliff.parallel.parmap2`.

    Warning
    -------
    This is implemented using ``multiprocessing.Queue``, which requires the function``f``
    to be picklable. If it is not the case (e.g. use KIM library functions), use
    :meth:`kliff.parallel.parmap2` that is based on ``multiprocessing.Pipe``.


    Example
    -------
    >>> def func(x, y, z=1):
    >>>     return x+y+z
    >>> X = range(3)
    >>> Y = range(3)
    >>> parmap1(func, X, 1, nprocs=2)  # [2,3,4]
    >>> parmap1(func, X, 1, 1, nprocs=2)  # [2,3,4]
    >>> parmap1(func, zip(X, Y), tuple_X=True, nprocs=2)  # [1,3,5]
    >>> parmap1(func, zip(X, Y), 1, tuple_X=True, nprocs=2)  # [1,3,5]
    """

    q_in = mp.Queue(nprocs)
    q_out = mp.Queue()

    processes = []
    for _ in range(nprocs):
        p = mp.Process(target=_func1, args=(f, q_in, q_out))
        p.daemon = True
        p.start()
        processes.append(p)

    N = 0
    for i, x in enumerate(X):
        N += 1
        if tuple_X:
            ix = (i, *x)
        else:
            ix = (i, x)
        q_in.put((ix, args))

    [q_in.put((None, None)) for _ in range(nprocs)]
    results = [q_out.get() for _ in range(N)]
    [p.join() for p in processes]

    return [r for i, r in sorted(results)]


def _func1(f, q_in, q_out):
    while True:
        ix, args = q_in.get()
        if ix is None:
            break
        i = ix[0]
        x = ix[1:]
        y = f(*x, *args)
        q_out.put((i, y))


def parmap2(f, X, *args, tuple_X=False, nprocs=mp.cpu_count()):
    """
    Parallelism over data.

    This is to mimic ``multiprocessing.Pool.map``, which requires the function ``f`` to be
    picklable. This function does not have this restriction and allows extra arguments to
    be used for the function ``f``.

    Parameters
    ----------
    f: function
        The function that operates on the data.

    X: list
        Data to be parallelized.

    args: args
        Extra positional arguments needed by the function ``f``.

    tuple_X: bool
        This depends on ``X``. It should be set to ``True`` if multiple arguments are
        parallelized and set to ``False`` if only one argument is parallelized. See
        ``Example`` below.

    nprocs: int
        Number of processors to use.

    Return
    ------
    list
        A list of results, corresponding to ``X``.

    Note
    ----
    This function is implemented using ``multiprocessing.Pipe``. The data is subdivided
    into ``nprocs`` groups and then each group of data is distributed to a process. The
    results from each group are then assembled together.  The data is shuffled to balance
    the load in each process.  See :meth:`kliff.parallel.parmap1` for another
    implementation that uses ``multiprocessing.Queue``.

    Example
    -------
    >>> def func(x, y, z=1):
    >>>     return x+y+z
    >>> X = range(3)
    >>> Y = range(3)
    >>> parmap2(func, X, 1, nprocs=2)  # [2,3,4]
    >>> parmap2(func, X, 1, 1, nprocs=2)  # [2,3,4]
    >>> parmap2(func, zip(X, Y), tuple_X=True, nprocs=2)  # [1,3,5]
    >>> parmap2(func, zip(X, Y), 1, tuple_X=True, nprocs=2)  # [1,3,5]
    """

    # shuffle and divide into nprocs equally-numbered parts
    if tuple_X:
        pairs = [(i, *x) for i, x in enumerate(X)]  # to make array_split work
    else:
        pairs = [(i, x) for i, x in enumerate(X)]
    random.shuffle(pairs)
    groups = np.array_split(pairs, nprocs)

    processes = []
    managers = []
    for i in range(nprocs):
        manager_end, worker_end = mp.Pipe(duplex=False)
        p = mp.Process(target=_func2, args=(f, groups[i], args, worker_end))
        p.daemon = True
        p.start()
        processes.append(p)
        managers.append(manager_end)
    results = []
    for m in managers:
        results.extend(m.recv())
    for p in processes:
        p.join()

    return [r for i, r in sorted(results)]


def _func2(f, iX, args, worker_end):
    results = []
    for ix in iX:
        i = ix[0]
        x = ix[1:]
        results.append((i, f(*x, *args)))
    worker_end.send(results)


def get_MPI_world_size():
    try:
        from mpi4py import MPI

        mpi4py_available = True
    except ImportError as e:
        mpi4py_available = False

    if mpi4py_available:
        return MPI.COMM_WORLD.Get_size()
    else:
        return 1



from typing import Dict, List, Tuple

import numpy as np
from kliff.atomic_data import atomic_number, atomic_species
from kliff.dataset.dataset import Configuration

from . import nl


class NeighborList:
    """
    Neighbor list class.

    This uses the same approach that `LAMMPS` and `KIM` adopt: The atoms in the
    configuration (assuming a total of N atoms) are named contributing atoms, and padding
    atoms are created to satisfy the boundary conditions. The contributing atoms are
    numbered as 1, 2, ... N-1, and the padding atoms are numbered as N, N+1, N+2...
    Neighbors of atom can include both contributing atoms and padding atoms.

    Args:
        conf: atomic configuration.
        infl_dist: Influence distance, within which atoms are interacting with each other.
            In literatures, this is usually referred as ``cutoff``.
        padding_need_neigh: Whether to generate neighbors for padding atoms.

    Attributes:
        coords: 2D array
            Coordinates of contributing and padding atoms.
        species: list
            Species string of contributing and padding atoms.
        image: 1D array
            Atom index, of which an atom is an image. The image of a contributing atom is
            itself.
        padding_coords: 2D array
            Coordinates of padding atoms.
        padding_species: list
            Species string and padding atoms.
        padding_image: 1D array
            Atom index, of which a padding atom is an image.

    Note:
        To get the total force on a contributing atom, the forces on all padding atoms
        that are images of the contributing atom should be added back to the contributing
        atom.
    """

    def __init__(
        self, conf: Configuration, infl_dist: float, padding_need_neigh: bool = False
    ):
        self.conf = conf
        self.infl_dist = infl_dist
        self.padding_need_neigh = padding_need_neigh

        # all atoms: contrib + padding
        self.coords = None
        self.species = None
        self.image = None

        self.padding_coords = None
        self.padding_species = None
        # padding_image[0] = 3: padding atom 1 is the image of contributing atom 3
        self.padding_image = None

        # neigh
        self.neigh = nl.initialize()
        self.create_neigh()

    def create_neigh(self):
        coords_cb = np.asarray(self.conf.coords, dtype=np.double)
        species_cb = self.conf.species
        cell = np.asarray(self.conf.cell, dtype=np.double)
        PBC = np.asarray(self.conf.PBC, dtype=np.intc)

        # create padding atoms
        species_code_cb = np.asarray(
            [atomic_number[s] for s in species_cb], dtype=np.intc
        )
        out = nl.create_paddings(self.infl_dist, cell, PBC, coords_cb, species_code_cb)
        coords_pd, species_code_pd, image_pd, error = out
        check_error(error, "nl.create_padding")
        species_pd = [atomic_species[i] for i in species_code_pd]

        self.padding_coords = np.asarray(coords_pd, dtype=np.double)
        self.padding_species = species_pd
        self.padding_image = np.asarray(image_pd, dtype=np.intc)

        num_cb = coords_cb.shape[0]
        num_pd = coords_pd.shape[0]

        self.coords = np.asarray(
            np.concatenate((coords_cb, coords_pd)), dtype=np.double
        )
        self.species = np.concatenate((species_cb, species_pd))
        self.image = np.asarray(
            np.concatenate((np.arange(num_cb), image_pd)), dtype=np.intc
        )
        # flag to indicate whether to create neighborlist for an atom
        need_neigh = np.ones(num_cb + num_pd, dtype=np.intc)
        if not self.padding_need_neigh:
            need_neigh[num_cb:] = 0

        # create neighbor list
        cutoffs = np.asarray([self.infl_dist], dtype=np.double)
        error = nl.build(self.neigh, self.coords, self.infl_dist, cutoffs, need_neigh)
        check_error(error, "nl.build")

    def get_neigh(self, index: int) -> Tuple[List[int], np.array, List[str]]:
        """
        Get the indices, coordinates, and species string of a given atom.

        Args:
            index: Atom number whose neighbor info is requested.

        Returns:
            neigh_indices: Indices of neighbor atoms in self.coords and self.species.
            neigh_coords: 2D array of shape (N, 3), where N is the number of neighbors.
                Coordinates of neighbor atoms.
            neigh_species: Species symbol of neighbor atoms.
        """

        cutoffs = np.asarray([self.infl_dist], dtype=np.double)
        neigh_list_index = 0
        num_neigh, neigh_indices, error = nl.get_neigh(
            self.neigh, cutoffs, neigh_list_index, index
        )
        check_error(error, "nl.get_neigh")

        neigh_coords = self.coords[neigh_indices]
        neigh_species = self.species[neigh_indices]

        return neigh_indices, neigh_coords, neigh_species

    def get_numneigh_and_neighlist_1D(
        self, request_padding: bool = False
    ) -> Tuple[np.array, np.array]:
        """
        Get the number of neighbors and neighbor list for all atoms.

        Args:
            request_padding: If ``True``, the returned number of neighbors and neighbor
                list include those for padding atoms; If ``False``, only return these
                for contributing atoms.

        Returns:
            numneigh: 1D array; number of neighbors for all atoms.
            neighlist: 1D array; indices of the neighbors for all atoms stacked into a
                1D array. Its total length is ``sum(numneigh)``, and the first
                ``numneigh[0]`` components are the neighbors of atom `0`, the next
                ``numneigh[1]`` components are the neighbors of atom `1` ....
        """
        if request_padding:
            if not self.padding_need_neigh:
                raise NeighborListError(
                    "Request to get neighbors of padding atoms, but "
                    '"padding_need_neigh" is set to "False" at initialization.'
                )
            N = len(self.coords)
        else:
            N = self.conf.get_num_atoms()

        cutoffs = np.asarray([self.infl_dist], dtype=np.double)
        neigh_list_index = 0

        numneigh = []
        neighlist = []
        for i in range(N):
            num_neigh, neigh_indices, error = nl.get_neigh(
                self.neigh, cutoffs, neigh_list_index, i
            )
            check_error(error, "nl.get_neigh")
            numneigh.append(num_neigh)
            neighlist.append(neigh_indices)
        neighlist = np.asarray(np.concatenate(neighlist), dtype=np.intc)
        numneigh = np.asarray(numneigh, dtype=np.intc)

        return numneigh, neighlist

    def get_coords(self) -> np.array:
        """
        Return coords of both contributing and padding atoms.
        Shape (N,3).
        """
        return self.coords.copy()

    def get_species(self) -> List[str]:
        """
        Return species of both contributing and padding atoms.
        """
        return self.species[:]

    def get_species_code(self, mapping: Dict[str, int]) -> np.array:
        """
        Integer species code of both contributing and padding atoms.

        Args:
            mapping: A mapping between species string and its code.

        Returns:
            1D array of integer species code.
        """
        return np.asarray([mapping[s] for s in self.species], dtype=np.intc)

    def get_image(self) -> np.array:
        """
        Return image of both contributing and padding atoms.

        It is a 1D array of atom index, of which an atom is an image.

        Note:
            The image of a contributing atom is itself.
        """
        return self.image.copy()

    def get_padding_coords(self) -> np.array:
        """
        Return coords of padding atoms, 2D array of shape (N,3), where N is the number
        of padding atoms.
        """
        return self.padding_coords.copy()

    def get_padding_speices(self) -> List[str]:
        """
        Return species string of padding atoms.
        """
        return self.padding_speices[:]

    def get_padding_species_code(self, mapping: Dict[str, int]) -> np.array:
        """
        Integer species code of padding atoms.

        Args:
            mapping: A mapping between species string and its code.

        Returns:
            1D array of integer species code for padding atoms.
        """
        return np.asarray([mapping[s] for s in self.padding_species], dtype=np.intc)

    def get_padding_image(self) -> np.array:
        """
        Return image of padding atoms.

        It is a 1D array of atom index, of which a padding atom is an image.

        """
        return self.padding_image.copy()

    def __del__(self):
        nl.clean(self.neigh)


def assemble_forces(forces: np.array, n: int, padding_image: np.array) -> np.array:
    """
    Assemble forces on padding atoms back to contributing atoms.

    Args:
        forces: Partial forces on both contributing and padding atoms. 2D array of shape
            (Nc+Np, 3). where Nc is the number of contributing atoms, and Np is the
            number of padding atoms. The first Nc rows are the forces for contributing
            atoms.
        n: Number of contributing atoms, i.e. Nc.
        padding_image: atom index, of which the padding atom is an image. 1D int array
            of shape (Np,).

    Returns:
        Total forces on contributing atoms. 2D array of shape (Nc, 3), where Nc is the
        number of contributing atoms.
    """

    # numpy slicing does not make a copy !!!
    total_forces = np.array(forces[:n])

    has_padding = True if padding_image.size != 0 else False

    if has_padding:

        pad_forces = forces[n:]
        n_padding = pad_forces.shape[0]

        if n < n_padding:
            for i in range(n):
                # indices of padding atoms that are images of contributing atom i
                indices = np.where(padding_image == i)
                total_forces[i] += np.sum(pad_forces[indices], axis=0)
        else:
            for f, org_index in zip(pad_forces, padding_image):
                total_forces[org_index] += f

    return total_forces


def assemble_stress(coords: np.array, forces: np.array, volume: float) -> np.array:
    """
    Calculate the virial stress using the negative f dot r method.

    Args:
        coords: Coordinates of both contributing and padding atoms. 2D array of shape
            (Nc+Np, 3), where Nc is the number of contributing atoms, and Np is the
            number of padding atoms. The first Nc rows are the coords of contributing
            atoms.
        forces: Partial forces on both contributing and padding atoms. 2D array of shape
            (Nc+Np, 3), where Nc is the number of contributing atoms, and Np is the
            number of padding atoms. The first Nc rows are the forces on contributing
            atoms.
        volume: Volume of the configuration.

    Return:
        Virial stress in Voigt notation. 1D array of shape (6,).
    """

    stress = np.zeros(6)
    stress[0] = -np.sum(np.multiply(coords[:, 0], forces[:, 0])) / volume
    stress[1] = -np.sum(np.multiply(coords[:, 1], forces[:, 1])) / volume
    stress[2] = -np.sum(np.multiply(coords[:, 2], forces[:, 2])) / volume
    stress[3] = -np.sum(np.multiply(coords[:, 1], forces[:, 2])) / volume
    stress[4] = -np.sum(np.multiply(coords[:, 2], forces[:, 0])) / volume
    stress[5] = -np.sum(np.multiply(coords[:, 0], forces[:, 1])) / volume

    return stress


class NeighborListError(Exception):
    def __init__(self, msg):
        super(NeighborListError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


def check_error(error, message=None):
    if error != 0 and error is not None:
        raise NeighborListError('Calling "{}" failed.'.format(message))



import logging
import warnings


def set_up_logger():
    logging.basicConfig(
        filename="kliff.log",
        format="%(asctime)s:%(name)s:%(levelname)s: %(message)s",
        level=logging.INFO,
    )


def log_entry(logger, message, level="info", print_end="\n", warning_category=Warning):
    logger_level = getattr(logger, level)
    logger_level(message)

    if level == "info":
        print(message, end=print_end)
    elif level == "warning":
        warnings.warn(message, category=warning_category)



import warnings

from kliff.log import set_up_logger
from kliff.utils import torch_available

__version__ = "0.2.0"

set_up_logger()

if not torch_available():
    warnings.warn(
        "'PyTorch' not found. All kliff machine learning modules (e.g. NeuralNetwork) "
        "are not imported. Ignore this if you want to use kliff to train "
        "physics-based models."
    )



from pathlib import Path

import numpy as np
import pytest
from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.models import KIMModel

ref_energies = [-277.409737571, -275.597759276, -276.528342759, -275.482988187]


ref_forces = [
    [
        [-1.15948917e-14, -1.15948917e-14, -1.16018306e-14],
        [-1.16989751e-14, -3.93232367e-15, -3.93232367e-15],
        [-3.92538477e-15, -1.17128529e-14, -3.92538477e-15],
    ],
    [
        [1.2676956, 0.1687802, -0.83520474],
        [-1.2536342, 1.19394052, -0.75371034],
        [-0.91847129, -0.26861574, -0.49488973],
    ],
    [
        [0.18042083, -0.11940541, 0.01800594],
        [0.50030564, 0.09165797, -0.09694234],
        [-0.23992404, -0.43625564, 0.14952855],
    ],
    [
        [-1.1114163, 0.21071302, 0.55246303],
        [1.51947195, -1.21522541, 0.7844481],
        [0.54684859, 0.01018317, 0.5062204],
    ],
]


class TestCalculator:
    def test_compute(self):
        test_file_path = Path(__file__).parents[1].joinpath("configs_extxyz")
        tset = Dataset(test_file_path.joinpath("Si_4"))
        configs = tset.get_configs()

        modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_005"
        model = KIMModel(modelname)

        # calculator
        calc = Calculator(model)
        compute_arguments = calc.create(configs)

        for i, ca in enumerate(compute_arguments):
            calc.compute(ca)
            energy = calc.get_energy(ca)
            forces = calc.get_forces(ca)[:3]

            assert energy == pytest.approx(ref_energies[i], 1e-6)
            assert np.allclose(forces, ref_forces[i])

    def test_parameter(self):
        modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_005"
        model = KIMModel(modelname)

        # parameters
        params = model.get_model_params()
        sigma = params["sigma"][0]
        A = params["A"][0]

        # optimizing parameters
        # B will not be optimized, only providing initial guess
        model.set_opt_params(
            sigma=[["default"]], B=[["default", "fix"]], A=[["default"]]
        )

        calc = Calculator(model)

        x0 = calc.get_opt_params()
        assert x0[0] == sigma
        assert x0[1] == A
        assert len(x0) == 2
        assert model.get_num_opt_params() == 2

        x1 = [i + 0.1 for i in x0]
        calc.update_model_params(x1)

        assert params["sigma"][0] == sigma + 0.1
        assert params["A"][0] == A + 0.1



import numpy as np
import pytest
from kliff.dataset import Dataset
from kliff.models.kim import KIMComputeArguments, KIMModel

ref_energies = [-277.409737571, -275.597759276, -276.528342759, -275.482988187]


ref_forces = [
    [
        [-1.15948917e-14, -1.15948917e-14, -1.16018306e-14],
        [-1.16989751e-14, -3.93232367e-15, -3.93232367e-15],
        [-3.92538477e-15, -1.17128529e-14, -3.92538477e-15],
    ],
    [
        [1.2676956, 0.1687802, -0.83520474],
        [-1.2536342, 1.19394052, -0.75371034],
        [-0.91847129, -0.26861574, -0.49488973],
    ],
    [
        [0.18042083, -0.11940541, 0.01800594],
        [0.50030564, 0.09165797, -0.09694234],
        [-0.23992404, -0.43625564, 0.14952855],
    ],
    [
        [-1.1114163, 0.21071302, 0.55246303],
        [1.51947195, -1.21522541, 0.7844481],
        [0.54684859, 0.01018317, 0.5062204],
    ],
]


class TestKIM:
    def test_compute(self):

        # model
        modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_005"
        model = KIMModel(modelname)

        # training set
        data = Dataset("./configs_extxyz/Si_4")
        configs = data.get_configs()

        # compute arguments
        compute_arguments = []
        for conf in configs:
            ca = model.create_a_kim_compute_argument()
            compute_arguments.append(
                KIMComputeArguments(
                    ca,
                    conf,
                    supported_species=model.get_supported_species(),
                    influence_distance=model.get_influence_distance(),
                )
            )
        for i, ca in enumerate(compute_arguments):
            ca.compute(model.kim_model)
            energy = ca.get_energy()
            forces = ca.get_forces()[:3]

            assert energy == pytest.approx(ref_energies[i], 1e-6)
            assert np.allclose(forces, ref_forces[i])

    def test_set_one_param(self):
        modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_005"
        model = KIMModel(modelname)

        # parameters
        params = model.get_model_params()

        # keep the original copy
        sigma = params["sigma"][0]

        model.set_one_opt_param(name="sigma", settings=[[sigma + 0.1]])

        assert params["sigma"][0] == sigma + 0.1

        # internal kim params
        kim_params = model.get_kim_model_params()
        assert kim_params["sigma"][0] == sigma + 0.1

    def test_set_opt_params(self):
        modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_005"
        model = KIMModel(modelname)

        # parameters
        params = model.get_model_params()
        sigma = params["sigma"][0]
        A = params["A"][0]

        model.set_opt_params(sigma=[[sigma + 0.1]], A=[[A + 0.1]])

        assert params["sigma"][0] == sigma + 0.1
        assert params["A"][0] == A + 0.1

        # internal kim params
        kim_params = model.get_kim_model_params()
        assert kim_params["sigma"][0] == sigma + 0.1
        assert kim_params["A"][0] == A + 0.1

    def test_get_update_params(self):
        modelname = "SW_StillingerWeber_1985_Si__MO_405512056662_005"
        model = KIMModel(modelname)

        # parameters
        params = model.get_model_params()
        sigma = params["sigma"][0]
        A = params["A"][0]

        # optimizing parameters
        # B will not be optimized, only providing initial guess
        model.set_opt_params(
            sigma=[["default"]], B=[["default", "fix"]], A=[["default"]]
        )

        x0 = model.get_opt_params()
        assert x0[0] == sigma
        assert x0[1] == A
        assert len(x0) == 2
        assert model.get_num_opt_params() == 2

        x1 = [i + 0.1 for i in x0]
        model.update_model_params(x1)

        assert params["sigma"][0] == sigma + 0.1
        assert params["A"][0] == A + 0.1

        # internal kim params
        kim_params = model.get_kim_model_params()
        assert kim_params["sigma"][0] == sigma + 0.1
        assert kim_params["A"][0] == A + 0.1



import torch
from torch.nn import *


# redefine Dropout layer
class Dropout(torch.nn.modules.dropout._DropoutNd):
    """
    A Dropout layer that zeros the same element of descriptor values for all atoms.

    Note `torch.nn.Dropout` dropout each component independently.

    Args:
        p: float
            probability of an element to be zeroed. Default: 0.5

        inplace: bool
            If set to `True`, will do this operation in-place. Default: `False`

    Shapes:
        Input: [N, D] or [1, N, D]
        Output: [N, D] or [1, N, D] (same as Input)
        The first dimension 1 is because the dataloader provides only sample each
        iteration.
    """

    def forward(self, input):
        dim = input.dim()
        shape = input.shape
        if dim == 2:
            shape_4D = (1, *shape, 1)
        elif dim == 3:
            if shape[0] != 1:
                raise Exception("Shape[0] needs to be 1 for a 3D tensor.")
            shape_4D = (*shape, 1)

        else:
            raise Exception(
                "Input need to be 2D or 3D tensor, but got a " "{}D tensor.".format(dim)
            )
        x = torch.reshape(input, shape_4D)
        x = torch.transpose(x, 1, 2)
        y = torch.nn.functional.dropout2d(x, self.p, self.training, self.inplace)
        y = torch.transpose(y, 1, 2)
        y = torch.reshape(y, shape)
        return y



import numpy as np
from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.loss import Loss
from kliff.models import KIMModel


def residual_fn(identifier, natoms, weight, prediction, reference, data):
    assert len(prediction) == 3 * natoms + 1
    return weight * (prediction - reference)


def init():
    model = KIMModel(model_name="SW_StillingerWeber_1985_Si__MO_405512056662_005")

    # Cannot set them all by calling this function only once, because the assertion
    # depends on order
    model.set_opt_params(A=[[5.0]])
    model.set_opt_params(B=[["default"]])
    model.set_opt_params(sigma=[[2.0951, "fix"]])
    model.set_opt_params(gamma=[[1.5]])

    tset = Dataset("./configs_extxyz/Si_4")
    configs = tset.get_configs()

    calc = Calculator(model)
    calc.create(configs, use_energy=True, use_forces=True)

    loss = Loss(calc, residual_fn=residual_fn, nprocs=1)

    return loss


def optimize(method, reference):
    loss = init()
    steps = 3
    result = loss.minimize(method, options={"disp": False, "maxiter": steps})
    assert np.allclose(result.x, reference)


def least_squares(method, reference):
    loss = init()
    steps = 3
    result = loss.minimize(method, max_nfev=steps, verbose=0)
    assert np.allclose(result.x, reference)


def test_lbfgsb():
    optimize("L-BFGS-B", [4.81469314, 1.52682165, 1.50169278])


def test_bfgs():
    optimize("BFGS", [4.33953474, 1.53585151, 2.36509375])


def test_cg():
    optimize("CG", [4.54044991, 1.562142, 5.16457532])


def test_powell():
    optimize("Powell", [0.09323494, 1.11123083, 2.10211015])


def test_lm():
    least_squares("lm", [6.42665872, 1.81078954, 1.93466249])


def test_trf():
    least_squares("trf", [6.42575061, 1.54254653, 2.13551637])


def test_dogbox():
    least_squares("dogbox", [6.42575107, 1.54254652, 2.13551639])



import logging

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from ..dataset.dataset_torch import FingerprintsDataset, fingerprints_collate_fn
from .model_torch import ModelTorch

logger = logging.getLogger(__name__)


class LinearRegression(ModelTorch):
    r"""Linear regression model."""

    def __init__(self, descriptor, seed=35):
        super(LinearRegression, self).__init__(descriptor, seed)

        desc_size = self.descriptor.get_size()
        self.layer = nn.Linear(desc_size, 1)

    def forward(self, x):
        return self.layer(x)

    def fit(self, path):
        """
        Fit the model using analytic solution.
        """
        fp = FingerprintsDataset(path)

        loader = DataLoader(
            dataset=fp, batch_size=1, collate_fn=fingerprints_collate_fn
        )

        X, y = self.prepare_data(loader)
        A = torch.inverse(torch.mm(X.t(), X))
        beta = torch.mv(torch.mm(A, X.t()), y)

        self.set_params(beta)

        msg = 'fit model "{}" finished.'.format(self.__class__.__name__)
        logger.info(msg)
        print(msg)

    def set_params(self, beta):
        r"""Set linear weight and bias.

        Parameters
        ----------
        beta: Tensor
            First component is bias and the remaining components is weight.
        """
        self.layer.weight = torch.nn.Parameter(beta[1:])
        self.layer.bias = torch.nn.Parameter(beta[0:1])

    def prepare_data(self, loader, use_energy=True, use_forces=False):
        X = []
        y = []
        for batch in loader:
            sample = batch[0]
            if use_energy:
                zeta = sample["zeta"]
                intercept = torch.ones(zeta.size()[0], 1)
                zeta = torch.cat((intercept, zeta), dim=1)

                # sum to get energy of the configuration; we can do this because the model
                # is linear
                zeta = torch.sum(zeta, 0, keepdim=True)  # 2D tensor
                e = torch.tensor([sample["energy"]])  # 1D tensor

            if use_forces:
                dzeta = sample["dzeta_dr"]  # 3D tensor (atom, desc, coords)
                # torch.zeros because derivative of intercept is 0
                intercept = torch.zeros(dzeta.size()[0], dzeta.size()[2])
                dzeta = torch.cat((intercept, dzeta), dim=1)

                dzeta = torch.sum(dzeta, 0)  # 2D tensor
                f = sample["forces"][0]  # 1D tensor

            if use_energy and use_forces:
                x_ = torch.cat((zeta, torch.transpose(dzeta)))
                y_ = torch.cat((e, f))
            elif use_energy:
                x_ = zeta
                y_ = e
            elif use_forces:
                x_ = torch.transpose(dzeta)
                y_ = f
            else:
                raise LinearRegressionError(
                    'Both "use_energy" and "use_forces" are "False".'
                )

            X.append(x_)
            y.append(y_)

        X = torch.cat(X)
        y = torch.cat(y)

        return X, y


class LinearRegressionError(Exception):
    def __init__(self, msg):
        super(LinearRegressionError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg



class InputError(Exception):
    def __init__(self, msg):
        super(InputError, self).__init__(msg)
        self.msg = msg


class KeyNotFoundError(Exception):
    def __init__(self, msg):
        super(KeyNotFoundError, self).__init__(msg)
        self.msg = msg


def report_import_error(package, classname=None):
    if classname is not None:
        msg = f"To use `{classname}`, `{package}` is need. Please install it first."
    else:
        msg = f"Package `{package}` needed by KLIFF not found. Please install it first."

    raise ImportError(msg)



from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import numpy as np
from kliff.error import InputError, KeyNotFoundError


def read_extxyz(
    filename: Path,
) -> Tuple[
    np.ndarray,
    List[str],
    np.ndarray,
    List[bool],
    Union[float, None],
    Union[np.ndarray, None],
    Union[List[float], None],
]:
    """
    Read atomic configuration stored in extended xyz file_format.

    Args:
        filename: filename to the extended xyz file

    Returns:
        cell: 3x3 array, supercell lattice vectors
        species: species of atoms
        coords: Nx3 array, coordinates of atoms
        PBC: periodic boundary conditions
        energy: potential energy of the configuration; `None` if not provided in file
        forces: Nx3 array, forces on atoms; `None` if not provided in file
        stress: 1D array of size 6, stress on the cell in Voigt notation; `None` if not
            provided in file
    """
    with open(filename, "r") as fin:
        lines = fin.readlines()

        try:
            natoms = int(lines[0].split()[0])
        except ValueError as e:
            raise InputError(f"{e}.\nCorrupted data at line 1 of file {filename}.")

        # lattice vector
        line = lines[1].replace("'", '"')
        cell = _parse_key_value(line, "Lattice", "float", 9, filename)
        cell = np.reshape(cell, (3, 3))

        # PBC
        PBC = _parse_key_value(line, "PBC", "int", 3, filename)

        # energy is optional
        try:
            in_quotes = _check_in_quotes(line, "Energy", filename)
            energy = _parse_key_value(line, "Energy", "float", 1, filename, in_quotes)[
                0
            ]
        except KeyNotFoundError:
            energy = None

        # stress is optional
        try:
            stress = _parse_key_value(line, "Stress", "float", 6, filename)
        except KeyNotFoundError:
            stress = None

        # body, species symbol, x, y, z (and fx, fy, fz if provided)
        species = []
        coords = []
        forces = []

        # if forces provided
        line = lines[2].strip().split()
        if len(line) == 4:
            has_forces = False
        elif len(line) == 7:
            has_forces = True
        else:
            raise InputError(f"Corrupted data at line 3 of file {filename}.")

        try:
            num_lines = 0
            for line in lines[2:]:
                num_lines += 1
                line = line.strip().split()
                if len(line) != 4 and len(line) != 7:
                    raise InputError(
                        f'Corrupted data at line {num_lines + 3} of file "{filename}".'
                    )
                if has_forces:
                    symbol, x, y, z, fx, fy, fz = line
                    species.append(symbol.lower().capitalize())
                    coords.append([float(x), float(y), float(z)])
                    forces.append([float(fx), float(fy), float(fz)])
                else:
                    symbol, x, y, z = line
                    species.append(symbol.lower().capitalize())
                    coords.append([float(x), float(y), float(z)])
        except ValueError as e:
            raise InputError(
                f"{e}.\nCorrupted data at line {num_lines + 3} of file {filename}."
            )

        if num_lines != natoms:
            raise InputError(
                f"Corrupted data file {filename}. Number of atoms is {natoms}, "
                f"whereas number of data lines is {num_lines}."
            )

        coords = np.asarray(coords)
        if has_forces:
            forces = np.asarray(forces)
        else:
            forces = None

        return cell, species, coords, PBC, energy, forces, stress


def write_extxyz(
    filename: Path,
    cell: np.ndarray,
    species: List[str],
    coords: np.ndarray,
    PBC: List[bool],
    energy: Optional[float] = None,
    forces: Optional[np.ndarray] = None,
    stress: Optional[List[float]] = None,
):
    """
    Write configuration info to a file in extended xyz file_format.

    Args:
        filename: filename to the extended xyz file
        cell: 3x3 array, supercell lattice vectors
        species: species of atoms
        coords: Nx3 array, coordinates of atoms
        PBC: periodic boundary conditions
        energy: potential energy of the configuration; If `None`, not write to file
        forces: Nx3 array, forces on atoms; If `None`, not write to file
        stress: 1D array of size 6, stress on the cell in Voigt notation; If `None`,
            not write to file
    """

    with open(filename, "w") as fout:

        # first line (number of atoms)
        natoms = len(species)
        fout.write("{}\n".format(natoms))

        # second line
        fout.write('Lattice="')
        for i, line in enumerate(cell):
            for j, item in enumerate(line):
                if i == 2 and j == 2:
                    fout.write('{:.15g}" '.format(item))
                else:
                    fout.write("{:.15g} ".format(item))

        PBC = [int(i) for i in PBC]
        fout.write('PBC="{} {} {}" '.format(PBC[0], PBC[1], PBC[2]))

        if energy is not None:
            fout.write('Energy="{:.15g}" '.format(energy))

        if stress is not None:
            fout.write('Stress="')
            for i, s in enumerate(stress):
                if i == 5:
                    fout.write('{:.15g}" '.format(s))
                else:
                    fout.write("{:.15g} ".format(s))

        properties = "Properties=species:S:1:pos:R:3"
        if forces is not None:
            properties += ":for:R:3\n"
        else:
            properties += "\n"
        fout.write(properties)

        # body
        for i in range(natoms):
            fout.write("{:2s} ".format(species[i]))
            fout.write(
                "{:23.15e} {:23.15e} {:23.15e} ".format(
                    coords[i][0], coords[i][1], coords[i][2]
                )
            )
            if forces is not None:
                fout.write(
                    "{:23.15e} {:23.15e} {:23.15e}".format(
                        forces[i][0], forces[i][1], forces[i][2]
                    )
                )
            fout.write("\n")


def _parse_key_value(
    line: str, key: str, dtype: str, size: int, filename: Path, in_quotes: bool = True
) -> List[Any]:
    """
    Given key, parse a string like ``other stuff key="value" other stuff`` to get value.

    If there is not space in value, the quotes `"` can be omitted.

    Args:
        line: The string line.
        key: Keyword to parse.
        dtype: Expected data type of value, `int` or `float`.
        size: Expected size of value.
        filename: File name where the line comes from.

    Returns:
        Values associated with key.
    """
    line = line.strip()
    key = _check_key(line, key, filename)
    try:
        value = line[line.index(key) :]
        if in_quotes:
            value = value[value.index('"') + 1 :]
            value = value[: value.index('"')]
        else:
            value = value[value.index("=") + 1 :]
            value = value.lstrip(" ")
            value += " "  # add an whitespace at end in case this is the last key
            value = value[: value.index(" ")]
        value = value.split()
    except Exception as e:
        raise InputError(f"{e}.\nCorrupted {key} data at line 2 of file {filename}.")

    if len(value) != size:
        raise InputError(
            f"Incorrect size of {key} at line 2 of file {filename};\n"
            f"required: {size}, provided: {len(value)}. Possibly, the quotes not match."
        )
    try:
        if dtype == "float":
            value = [float(i) for i in value]
        elif dtype == "int":
            value = [int(i) for i in value]
    except Exception as e:
        raise InputError(f"{e}.\nCorrupted {key} data at line 2 of file {filename}.")

    return value


def _check_key(line, key, filename):
    """
    Check whether a key or its lowercase counter part is in line.
    """
    if key not in line:
        key_lower = key.lower()
        if key_lower not in line:
            raise KeyNotFoundError(f"{key} not found at line 2 of file {filename}.")
        else:
            key = key_lower
    return key


def _check_in_quotes(line, key, filename):
    """
    Check whether ``key=value`` or ``key="value"`` in line.
    """
    key = _check_key(line, key, filename)
    value = line[line.index(key) :]
    value = value[value.index("=") + 1 :]
    value = value.lstrip(" ")
    if value[0] == '"':
        return True
    else:
        return False



"""
Continue the training of example_nn_Si.py.

Suppose you've run ``example_nn_Si.py`` and saved the model; here we continue training
the model for more steps.

Most of the settings are the same as ``example_nn_Si.py``. We only need to load the
parameters of the model and the state dictionary of the parameter.
"""

from kliff import nn
from kliff.calculators import CalculatorTorch
from kliff.dataset import Dataset
from kliff.descriptors import SymmetryFunction
from kliff.loss import Loss
from kliff.models import NeuralNetwork
from kliff.utils import download_dataset

descriptor = SymmetryFunction(
    cut_name="cos", cut_dists={"Si-Si": 5.0}, hyperparams="set51", normalize=True
)

N1 = 10
N2 = 10
model = NeuralNetwork(descriptor)
model.add_layers(
    # first hidden layer
    nn.Linear(descriptor.get_size(), N1),
    nn.Tanh(),
    # second hidden layer
    nn.Linear(N1, N2),
    nn.Tanh(),
    # output layer
    nn.Linear(N2, 1),
)
model.set_save_metadata(prefix="./kliff_saved_model", start=5, frequency=2)

##########################################################################################
# Load the parameters from the saved model.
# If we are load a model to continue the training (the case here), ``mode`` needs to be
# set to ``train``; if we load the model for evaluation, it should be ``eval``. For
# fully-connected layer, this actually does not matter. But for dropout and batchNorm
# layers, the two modes are different.
model.load("final_model.pkl", mode="train")


# training set
dataset_path = download_dataset(dataset_name="Si_training_set")
dataset_path = dataset_path.joinpath("varying_alat")
tset = Dataset(dataset_path)
configs = tset.get_configs()

# calculator
calc = CalculatorTorch(model)
calc.create(configs, reuse=True)


# loss
loss = Loss(calc, residual_data={"forces_weight": 0.3})

##########################################################################################
# load the state dictionary of the optimizer.
# We also set ``start_epoch`` to ``10`` such that the epoch number continues from the last
# training.

loss.load_optimizer_state("optimizer_stat.pkl")
result = loss.minimize(
    method="Adam", num_epochs=10, start_epoch=10, batch_size=100, lr=0.001
)



import numpy as np
from kliff.dataset.dataset import Configuration
from kliff.neighbor import NeighborList

target_coords = np.asarray(
    [
        [0.000000e00, 0.000000e00, 0.000000e00],
        [1.234160e00, 7.125400e-01, 0.000000e00],
        [0.000000e00, 0.000000e00, 3.355150e00],
        [1.234160e00, 7.125400e-01, 3.355150e00],
        [-2.468323e00, -1.425090e00, 0.000000e00],
        [-2.468323e00, -1.425090e00, 3.355150e00],
        [-1.234162e00, 7.125400e-01, 0.000000e00],
        [-1.234162e00, 7.125400e-01, 3.355150e00],
        [-1.000000e-06, -1.425090e00, 0.000000e00],
        [-1.000000e-06, -1.425090e00, 3.355150e00],
        [1.234161e00, 2.137630e00, 0.000000e00],
        [1.234161e00, 2.137630e00, 3.355150e00],
        [2.468322e00, 0.000000e00, 0.000000e00],
        [2.468322e00, 0.000000e00, 3.355150e00],
        [3.702483e00, 2.137630e00, 0.000000e00],
        [3.702483e00, 2.137630e00, 3.355150e00],
    ]
)

target_species = ["C"] * 16
target_species[0] = "O"
target_species[10] = "O"
target_species[12] = "O"
target_species[14] = "O"

all_indices = [[6, 1, 8], [0, 10, 12], [7, 3, 9], [2, 11, 13]]
all_numneigh = [len(i) for i in all_indices]


def test_neigh():
    conf = Configuration.from_file(
        "configs_extxyz/bilayer_graphene/bilayer_sep3.36_i0_j0.xyz"
    )
    conf.species[0] = "O"

    neigh = NeighborList(conf, infl_dist=2, padding_need_neigh=False)
    coords = neigh.get_coords()
    species = neigh.get_species()

    assert np.allclose(coords, target_coords)
    assert np.array_equal(species, target_species)

    # contributing
    for i in range(conf.get_num_atoms()):
        nei_indices, nei_coords, nei_species = neigh.get_neigh(i)
        assert np.allclose(nei_indices, all_indices[i])

    # padding
    for i in range(conf.get_num_atoms(), len(coords)):
        nei_indices, nei_coords, nei_species = neigh.get_neigh(i)
        assert nei_indices.size == 0
        assert nei_coords.size == 0
        assert nei_species.size == 0

    numneigh, neighlist = neigh.get_numneigh_and_neighlist_1D(request_padding=False)
    np.array_equal(numneigh, all_numneigh)
    np.array_equal(neighlist, np.concatenate(all_indices))



import logging
import multiprocessing as mp
import os
from pathlib import Path
from typing import Dict, List, Optional

import torch
import torch.distributed as dist
from kliff.dataset.dataset import Configuration
from kliff.dataset.dataset_torch import FingerprintsDataset, fingerprints_collate_fn
from kliff.models.model_torch import ModelTorch
from kliff.models.neural_network import NeuralNetwork
from torch.nn.parallel import DistributedDataParallel
from torch.utils.data import DataLoader

logger = logging.getLogger(__name__)


class CalculatorTorch:
    """
    A calculator for torch based models.

    Args:
        model: torch models, e.g. :class:`~kliff.neuralnetwork.NeuralNetwork`.
    """

    implemented_property = ["energy", "forces", "stress"]

    def __init__(self, model: ModelTorch):

        self.model = model
        self.dtype = self.model.descriptor.dtype
        self.fingerprints_path = None

        self.use_energy = None
        self.use_forces = None
        self.use_stress = None

        self.results = dict([(i, None) for i in self.implemented_property])

    def create(
        self,
        configs: List[Configuration],
        use_energy: bool = True,
        use_forces: bool = True,
        use_stress: bool = False,
        fingerprints_path: Optional[Path] = None,
        fingerprints_mean_and_stdev_path: Optional[Path] = None,
        reuse: bool = False,
        serial: bool = False,
        nprocs: int = mp.cpu_count(),
    ):
        """
        Process configs to generate fingerprints.

        Args:
            configs: atomic configurations
            use_energy: Whether to require the calculator to compute energy.
            use_forces: Whether to require the calculator to compute forces.
            use_stress: Whether to require the calculator to compute stress.
            fingerprints_path: Path to the to be generated fingerprints. If ``None``,
            default to ``./fingerprint.pkl``.
            fingerprints_mean_and_stdev_path: Path to the mean and standard deviation of
                the fingerprints. If ``normalize`` is not required by a descriptor,
                this is ignored. Otherwise, the mean and standard deviation read from
                ``fingerprints_mean_and_stdev_path``, are used to normalize the
                fingerprints. If ``None``, mean and standard deviation will be calculated
                from the descriptors and write to ``./fingerprints_mean_and_stdev.pkl``;
            reuse: If ``True``, reuse the fingerprints if found existing one. Otherwise,
                generate fingerprints from scratch no matter there is existing one or not.
            serial: Compute fingerprints in serial mode. Memory efficient.
            nprocs: Number of processes to use to generate the fingerprints.
                If ``serial`` is ``True``, this is ignored.
        """

        self.configs = configs
        self.use_energy = use_energy
        self.use_forces = use_forces
        self.use_stress = use_stress

        if isinstance(configs, Configuration):
            configs = [configs]

        # generate pickled fingerprints
        self.fingerprints_path = self.model.descriptor.generate_fingerprints(
            configs,
            use_forces,
            use_stress,
            reuse,
            fingerprints_path,
            fingerprints_mean_and_stdev_path,
            serial,
            nprocs,
        )

    def get_compute_arguments(self, batch_size: int = 1):
        """
        Return the dataloader with batch size set to ``batch_size``.
        """
        fname = self.fingerprints_path
        fp = FingerprintsDataset(fname)
        loader = DataLoader(
            dataset=fp, batch_size=batch_size, collate_fn=fingerprints_collate_fn
        )

        return loader

    def fit(self):
        path = self.fingerprints_path
        self.model.fit(path)

    def compute(self, batch):

        grad = self.use_forces or self.use_stress

        # collate batch input to NN
        zeta_config = [sample["zeta"] for sample in batch]
        if grad:
            for zeta in zeta_config:
                zeta.requires_grad_(True)

        # evaluate model
        zeta_stacked = torch.cat(zeta_config, dim=0)
        energy_atom = self.model(zeta_stacked)

        # energy
        natoms_config = [len(zeta) for zeta in zeta_config]
        energy_config = [e.sum() for e in torch.split(energy_atom, natoms_config)]

        # forces and stress
        if not self.use_forces:
            forces_config = None
        else:
            forces_config = []
        if not self.use_stress:
            stress_config = None
        else:
            stress_config = []
        if grad:
            for i, sample in enumerate(batch):

                # derivative of energy w.r.t. zeta
                energy = energy_config[i]
                zeta = zeta_config[i]
                dedz = torch.autograd.grad(energy, zeta, create_graph=True)[0]
                zeta.requires_grad_(False)  # no need of grad any more

                if self.use_forces:
                    dzetadr_forces = sample["dzetadr_forces"]
                    f = self.compute_forces(dedz, dzetadr_forces)
                    forces_config.append(f)

                if self.use_stress:
                    dzetadr_stress = sample["dzetadr_stress"]
                    volume = sample["dzetadr_volume"]
                    s = self.compute_stress(dedz, dzetadr_stress, volume)
                    stress_config.append(s)

        self.results["energy"] = energy_config
        self.results["forces"] = forces_config
        self.results["stress"] = stress_config
        return {
            "energy": energy_config,
            "forces": forces_config,
            "stress": stress_config,
        }

    @staticmethod
    def compute_forces(denergy_dzeta, dzetadr):
        forces = -torch.tensordot(denergy_dzeta, dzetadr, dims=([0, 1], [0, 1]))
        return forces

    @staticmethod
    def compute_stress(denergy_dzeta, dzetadr, volume):
        forces = torch.tensordot(denergy_dzeta, dzetadr, dims=([0, 1], [0, 1])) / volume
        return forces

    def get_energy(self, batch):
        return self.results["energy"]

    def get_forces(self, batch):
        return self.results["forces"]

    def get_stress(self, batch):
        return self.results["stress"]


class CalculatorTorchSeparateSpecies(CalculatorTorch):
    """
    A calculator supporting models of difference species.

    Args:
        models: {species:model} with species specifying the chemical symbol for the model.
    """

    def __init__(
        self,
        models: Dict[str, NeuralNetwork],
    ):
        self.models = models

        self.dtype = None
        for s, m in self.models.items():
            if self.dtype is None:
                self.dtype = m.descriptor.dtype
            else:
                if self.dtype != m.descriptor.dtype:
                    raise CalculatorTorchError("inconsistent `dtype` from descriptors.")

        # TODO change this (we now temporarily set model to the last one)
        self.model = m

        self.fingerprints_path = None

        self.use_energy = None
        self.use_forces = None
        self.use_stress = None

        self.results = dict([(i, None) for i in self.implemented_property])

    def compute(self, batch):

        grad = self.use_forces or self.use_stress

        # collate batch by species
        zeta_config = [sample["zeta"] for sample in batch]
        if grad:
            for zeta in zeta_config:
                zeta.requires_grad_(True)

        supported_species = self.models.keys()
        zeta_by_species = {s: [] for s in supported_species}
        config_id_by_species = {s: [] for s in supported_species}
        zeta_config = []

        for i, sample in enumerate(batch):
            zeta = sample["zeta"]
            species = sample["configuration"].species
            zeta.requires_grad_(True)
            zeta_config.append(zeta)

            for s, z in zip(species, zeta):
                # TODO move check to somewhere else to speed up computation
                if s not in supported_species:
                    raise CalculatorTorchError(f"No model for species: {s}")
                else:
                    zeta_by_species[s].append(z)
                    config_id_by_species[s].append(i)

        # evaluate model to compute energy
        energy_config = [None for _ in range(len(batch))]
        for s, zeta in zeta_by_species.items():

            # have no species "s" in this batch of data
            if not zeta:  # zeta == []
                continue

            z_tensor = torch.stack(zeta)  # convert a list of tensor to tensor
            energy = self.models[s](z_tensor)

            for e_atom, i in zip(energy, config_id_by_species[s]):
                if energy_config[i] is None:
                    energy_config[i] = e_atom
                else:
                    # note cannot use +=, energy e_atom is a view
                    energy_config[i] = energy_config[i] + e_atom

        # forces and stress
        if not self.use_forces:
            forces_config = None
        else:
            forces_config = []
        if not self.use_stress:
            stress_config = None
        else:
            stress_config = []
        if grad:
            for i, sample in enumerate(batch):

                # derivative of energy w.r.t. zeta
                energy = energy_config[i]
                zeta = zeta_config[i]
                dedz = torch.autograd.grad(energy, zeta, create_graph=True)[0]
                zeta.requires_grad_(False)  # no need of grad any more

                if self.use_forces:
                    dzetadr_forces = sample["dzetadr_forces"]
                    f = self.compute_forces(dedz, dzetadr_forces)
                    forces_config.append(f)

                if self.use_stress:
                    dzetadr_stress = sample["dzetadr_stress"]
                    volume = sample["dzetadr_volume"]
                    s = self.compute_stress(dedz, dzetadr_stress, volume)
                    stress_config.append(s)

        self.results["energy"] = energy_config
        self.results["forces"] = forces_config
        self.results["stress"] = stress_config
        return {
            "energy": energy_config,
            "forces": forces_config,
            "stress": stress_config,
        }


class CalculatorTorchDDPCPU(CalculatorTorch):
    def __init__(self, model, rank, world_size):
        super(CalculatorTorchDDPCPU, self).__init__(model)
        self.set_up(rank, world_size)

    def set_up(self, rank, world_size):
        os.environ["MASTER_ADDR"] = "localhost"
        os.environ["MASTER_PORT"] = "12355"
        dist.init_process_group("gloo", rank=rank, world_size=world_size)

    def clean_up(self):
        dist.destroy_process_group()

    def compute(self, batch):
        grad = self.use_forces

        # collate batch input to NN
        zeta_config = self._collate(batch, "zeta")
        if grad:
            for zeta in zeta_config:
                zeta.requires_grad_(True)
        zeta_stacked = torch.cat(zeta_config, dim=0)

        # evaluate model
        model = DistributedDataParallel(self.model)
        energy_atom = model(zeta_stacked)

        # energy
        natoms_config = [len(zeta) for zeta in zeta_config]
        energy_config = [e.sum() for e in torch.split(energy_atom, natoms_config)]

        # forces
        if grad:
            dzetadr_config = self._collate(batch, "dzetadr")
            forces_config = self.compute_forces_config(
                energy_config, zeta_config, dzetadr_config
            )
            for zeta in zeta_config:
                zeta.requires_grad_(False)
        else:
            forces_config = None

        return {"energy": energy_config, "forces": forces_config}

    def __del__(self):
        self.clean_up()


class CalculatorTorchError(Exception):
    def __init__(self, msg):
        super(CalculatorTorchError, self).__init__(msg)
        self.msg = msg



from kliff.calculators.calculator import Calculator, _WrapperCalculator
from kliff.dataset import Dataset
from kliff.loss import Loss

# training set
tset = Dataset("Si_training_set")
configs = tset.get_configs()

# calculators
calc1 = Calculator(model="Three_Body_Stillinger_Weber_Si__MO_405512056662_004")
calc1.create(configs)
calc1.set_fitting_params(A=[[16.0, 1.0, 20]], B=[["DEFAULT"]])
calc2 = Calculator(model="Three_Body_Stillinger_Weber_Si__MO_405512056662_004")
calc2.create(configs)
calc2.set_fitting_params(sigma=[[2.0951, "fix"]], gamma=[[1.1]])

# wrapper
calc = _WrapperCalculator(calc1, calc2)

# loss
with Loss(calc, nprocs=2) as loss:
    result = loss.minimize(method="L-BFGS-B", options={"disp": True, "maxiter": 10})



import logging
import os
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence

import numpy as np
from kliff.dataset.dataset import Configuration
from kliff.error import report_import_error
from kliff.log import log_entry
from kliff.models.model import ComputeArguments, Model
from kliff.models.parameter import Parameter
from kliff.neighbor import assemble_forces, assemble_stress

try:
    import kimpy
    from kimpy import neighlist as nl

    kimpy_avail = True
except ImportError:
    kimpy_avail = False


logger = logging.getLogger(__name__)


class KIMComputeArguments(ComputeArguments):
    """
    KIMModel potentials arguments.

    Args:
        kim_ca: KIM compute argument, can be created by
            :meth:`~kliff.models.KIMModels.create_a_compute_argument()`.
        conf: atomic configurations
        supported_species: species supported by the potential model, with chemical
            symbol as key and integer code as value.
        influence_distance: influence distance (aka cutoff distance) to calculate neighbors
        compute_energy: whether to compute energy
        compute_forces: whether to compute forces
        compute_stress: whether to compute stress
    """

    implemented_property = []

    def __init__(
        self,
        kim_ca,
        config: Configuration,
        supported_species: Dict[str, int],
        influence_distance: float,
        compute_energy: bool = True,
        compute_forces: bool = True,
        compute_stress: bool = False,
    ):
        if not kimpy_avail:
            report_import_error("kimpy", self.__class__.__name__)

        # kim compute argument
        self.kim_ca = kim_ca

        # get supported property
        self._get_implemented_property()

        super(KIMComputeArguments, self).__init__(
            config,
            supported_species,
            influence_distance,
            compute_energy,
            compute_forces,
            compute_stress,
        )

        # neighbor list
        self.neigh = None
        self.num_contributing_particles = None
        self.num_padding_particles = None
        self.padding_image_of = None

        # model input
        self.num_particles = None
        self.species_code = None
        self.particle_contributing = None
        self.coords = None

        # model output
        self.energy = None
        self.forces = None

        self._init_neigh()
        self._update_neigh(influence_distance)
        self._register_data(compute_energy, compute_forces)

    def _get_implemented_property(self):
        """ "
        Get implemented property of model.
        """

        # check compute arguments
        kim_can = kimpy.compute_argument_name
        N = kim_can.get_number_of_compute_argument_names()

        for i in range(N):
            name, error = kim_can.get_compute_argument_name(i)
            check_error(error, "kim_can.get_compute_argument_name")

            dtype, error = kim_can.get_compute_argument_data_type(name)
            check_error(error, "kim_can.get_compute_argument_data_type")

            support_status, error = self.kim_ca.get_argument_support_status(name)
            check_error(error, "kim_ca.get_argument_support_status")

            # calculator can only handle energy and forces
            if support_status == kimpy.support_status.required:
                if name != kim_can.partialEnergy and name != kim_can.partialForces:
                    report_error(f"Unsupported required ComputeArgument `{name}`")

            # supported property
            if name == kim_can.partialEnergy:
                self.implemented_property.append("energy")
            elif name == kim_can.partialForces:
                self.implemented_property.append("forces")
                self.implemented_property.append("stress")

        # check compute callbacks
        kim_ccn = kimpy.compute_callback_name
        num_callbacks = kim_ccn.get_number_of_compute_callback_names()

        for i in range(num_callbacks):
            name, error = kim_ccn.get_compute_callback_name(i)
            check_error(error, "kim_ccn.get_compute_callback_name")

            support_status, error = self.kim_ca.get_callback_support_status(name)
            check_error(error, "compute_arguments.get_callback_support_status")

            # calculator only provides get_neigh
            if support_status == kimpy.support_status.required:
                if name != kim_ccn.GetNeighborList:
                    report_error(f"Unsupported required ComputeCallback `{name}`")

    def _init_neigh(self):

        # create neighbor list
        neigh = nl.initialize()
        self.neigh = neigh

        # register get neigh callback
        error = self.kim_ca.set_callback_pointer(
            kimpy.compute_callback_name.GetNeighborList, nl.get_neigh_kim(), neigh
        )
        check_error(error, "compute_arguments.set_callback_pointer")

    def _update_neigh(self, influence_distance: float):
        """
        Update neighbor list and model input.
        """

        # inquire information from conf
        cell = np.asarray(self.conf.cell, dtype=np.double)
        PBC = np.asarray(self.conf.PBC, dtype=np.intc)
        contributing_coords = np.asarray(self.conf.coords, dtype=np.double)
        contributing_species = self.conf.species
        num_contributing = self.conf.get_num_atoms()
        self.num_contributing_particles = num_contributing

        # species support and code
        unique_species = list(set(contributing_species))
        species_map = dict()
        for s in unique_species:
            if s in self.supported_species:
                species_map[s] = self.supported_species[s]
            else:
                report_error(f"species `{s}` not supported by model")
        contributing_species_code = np.array(
            [species_map[s] for s in contributing_species], dtype=np.intc
        )

        if any(PBC):  # need padding atoms
            out = nl.create_paddings(
                influence_distance,
                cell,
                PBC,
                contributing_coords,
                contributing_species_code,
            )
            padding_coords, padding_species_code, self.padding_image_of, error = out
            check_error(error, "nl.create_paddings")

            num_padding = padding_species_code.size
            self.num_particles = np.array(
                [num_contributing + num_padding], dtype=np.intc
            )
            tmp = np.concatenate((contributing_coords, padding_coords))
            self.coords = np.asarray(tmp, dtype=np.double)
            tmp = np.concatenate((contributing_species_code, padding_species_code))
            self.species_code = np.asarray(tmp, dtype=np.intc)
            self.particle_contributing = np.ones(self.num_particles[0], dtype=np.intc)
            self.particle_contributing[num_contributing:] = 0

            # TODO check whether padding need neigh and create accordingly
            # for now, create neigh for all atoms, including paddings
            need_neigh = np.ones(self.num_particles[0], dtype=np.intc)

        else:  # do not need padding atoms
            self.padding_image_of = np.array([])
            self.num_particles = np.array([num_contributing], dtype=np.intc)
            self.coords = np.array(contributing_coords, dtype=np.double)
            self.species_code = np.array(contributing_species_code, dtype=np.intc)
            self.particle_contributing = np.ones(num_contributing, dtype=np.intc)
            need_neigh = self.particle_contributing

        error = nl.build(
            self.neigh,
            self.coords,
            influence_distance,
            np.asarray([influence_distance], dtype=np.double),
            need_neigh,
        )
        check_error(error, "nl.build")

    def _register_data(self, compute_energy=True, compute_forces=True):
        """
        Register model input and output data in KIM API.
        """

        # check whether model support energy and forces
        kim_can = kimpy.compute_argument_name
        if compute_energy:
            name = kim_can.partialEnergy
            support_status, error = self.kim_ca.get_argument_support_status(name)
            check_error(error, "kim_ca.get_argument_support_status")
            if not (
                support_status == kimpy.support_status.required
                or support_status == kimpy.support_status.optional
            ):
                report_error("Energy not supported by model")

        if compute_forces:
            name = kim_can.partialForces
            support_status, error = self.kim_ca.get_argument_support_status(name)
            check_error(error, "kim_ca.get_argument_support_status")
            if not (
                support_status == kimpy.support_status.required
                or support_status == kimpy.support_status.optional
            ):
                report_error("Forces not supported by model")

        # register argument
        error = self.kim_ca.set_argument_pointer(
            kim_can.numberOfParticles, self.num_particles
        )
        check_error(error, "kim_can.set_argument_pointer")

        error = self.kim_ca.set_argument_pointer(
            kim_can.particleSpeciesCodes, self.species_code
        )
        check_error(error, "kim_can.set_argument_pointer")

        error = self.kim_ca.set_argument_pointer(
            kim_can.particleContributing, self.particle_contributing
        )
        check_error(error, "kim_can.set_argument_pointer")

        error = self.kim_ca.set_argument_pointer(kim_can.coordinates, self.coords)
        check_error(error, "kim_can.set_argument_pointer")

        if compute_energy:
            self.energy = np.array([0.0], dtype=np.double)
            error = self.kim_ca.set_argument_pointer(kim_can.partialEnergy, self.energy)
            check_error(error, "kim_can.set_argument_pointer")
        else:
            self.energy = None
            error = self.kim_ca.set_argument_null_pointer(kim_can.partialEnergy)
            check_error(error, "kim_can.set_argument_null_pointer")

        if compute_forces:
            self.forces = np.zeros([self.num_particles[0], 3], dtype=np.double)
            error = self.kim_ca.set_argument_pointer(kim_can.partialForces, self.forces)
            check_error(error, "kim_can.set_argument_pointer")
        else:
            self.forces = None
            error = self.kim_ca.set_argument_null_pointer(kim_can.partialForces)
            check_error(error, "kim_can.set_argument_null_pointer")

    def compute(self, kim_model):
        error = kim_model.compute(self.kim_ca)
        check_error(error, "kim_model.compute")

        if self.compute_energy:
            self.results["energy"] = self.energy[0]
        if self.compute_forces:
            forces = assemble_forces(
                self.forces, self.num_contributing_particles, self.padding_image_of
            )
            self.results["forces"] = forces
        if self.compute_stress:
            volume = self.conf.get_volume()
            stress = assemble_stress(self.coords, self.forces, volume)
            self.results["stress"] = stress

    def __del__(self):
        """
        Garbage collection to destroy the neighbor list automatically.
        """
        if self.neigh is not None:
            nl.clean(self.neigh)
            self.neigh = None


class KIMModel(Model):
    """
    A general interface to any KIM model.

    Args:
        model_name: name of a KIM model. Available models can be found at:
            https://openkim.org.
            For example `SW_StillingerWeber_1985_Si__MO_405512056662_005`.
        params_relation_callback: A callback function to set the relations between
            parameters, which are called each minimization step after the optimizer
            updates the parameters. The function with be given a dictionary of
            :meth:`~kliff.model.parameter.Parameter` as argument, which can
            then be manipulated to set relations between parameters.

            Example:

            In the following example, we set the value of B[0] to 2 * A[0].

            def params_relation(model_params):
                A = model_params['A']
                B = model_params['B']
                B[0] = 2*A[0]
    """

    def __init__(
        self,
        model_name: str,
        params_relation_callback: Optional[Callable] = None,
    ):
        if not kimpy_avail:
            report_import_error("kimpy", self.__class__.__name__)

        self.kim_model = self._create_kim_model(model_name)

        super(KIMModel, self).__init__(model_name, params_relation_callback)

    def init_model_params(self) -> Dict[str, Parameter]:
        return self.get_kim_model_params()

    def init_influence_distance(self) -> float:
        return self.kim_model.get_influence_distance()

    def init_supported_species(self) -> Dict[str, int]:
        species = {}
        num_kim_species = kimpy.species_name.get_number_of_species_names()

        for i in range(num_kim_species):
            species_name, error = kimpy.species_name.get_species_name(i)
            check_error(error, "kimpy.species_name.get_species_name")
            supported, code, error = self.kim_model.get_species_support_and_code(
                species_name
            )
            check_error(error, "kim_model.get_species_support_and_code")
            if supported:
                species[str(species_name)] = code

        return species

    def get_compute_argument_class(self):
        return KIMComputeArguments

    @staticmethod
    def _create_kim_model(model_name: str):
        """
        Create a new kim model.
        """
        units_accepted, model, error = kimpy.model.create(
            kimpy.numbering.zeroBased,
            kimpy.length_unit.A,
            kimpy.energy_unit.eV,
            kimpy.charge_unit.e,
            kimpy.temperature_unit.K,
            kimpy.time_unit.ps,
            model_name,
        )
        check_error(error, "kimpy.model.create")
        if not units_accepted:
            report_error("requested units not accepted in kimpy.model.create")
        return model

    def get_kim_model_params(self) -> Dict[str, Parameter]:
        """
        Inquire the KIM model to get all the parameters.

        Returns:
            {name, parameter}, all parameters in a kim model.
        """
        params = dict()

        num_params = self.kim_model.get_number_of_parameters()
        for i in range(num_params):
            out = self.kim_model.get_parameter_metadata(i)
            dtype, extent, name, description, error = out
            check_error(error, "model.get_parameter_data_type_extent_and_description")

            values = []
            for j in range(extent):
                if str(dtype) == "Double":
                    val, error = self.kim_model.get_parameter_double(i, j)
                    check_error(error, "model.get_parameter_double")
                    values.append(val)
                elif str(dtype) == "Int":
                    val, error = self.kim_model.get_parameter_int(i, j)
                    check_error(error, "model.get_parameter_int")
                    values.append(val)
                else:  # should never reach here
                    report_error(f"get unexpected parameter data type `{dtype}`")

                params[name] = Parameter(value=values, index=i)

        return params

    def create_a_kim_compute_argument(self):
        """
        Create a compute argument for the KIM model.
        """
        kim_ca, error = self.kim_model.compute_arguments_create()
        check_error(error, "kim_model.compute_arguments_create")

        return kim_ca

    def set_opt_params(self, **kwargs):
        """
        Set the parameters that will be optimized.

        One or more parameters can be set. Each argument is for one parameter, where the
        argument name is the parameter name, the value of the argument is the
        settings(including initial value, fix flag, lower bound, and upper bound).

        The value of the argument should be a list of list, where each inner list is for
        one component of the parameter, which can contain 1, 2, or 3 elements.
         See `~kliff.model.model.Model.read_opt_params()` for the options of the elements.

        Example:
           instance.set(A=[['DEFAULT'], [2.0, 1.0, 3.0]], B=[[1.0, 'FIX'], [2.0, 'INF', 3.0]])
        """
        self.opt_params.set(**kwargs)

        # update kim internal model param (note, set_one will update model_params)
        for name, _ in kwargs.items():
            p_idx = self.model_params[name].index
            for c_idx, v in enumerate(self.model_params[name].value):
                self.kim_model.set_parameter(p_idx, c_idx, v)

        self.kim_model.clear_then_refresh()

        # reset influence distance in case it changes
        self.init_influence_distance()

    def set_one_opt_param(self, name: str, settings: List[List[Any]]):
        """
        Set one parameter that will be optimized.

        The name of the parameter should be given as the first entry of a list (or tuple),
        and then each data line should be given in in a list.

        Args:
            name: name of a fitting parameter
            settings: initial value, flag to fix a parameter, lower and upper bounds of a
                parameter.

        Example:
            name = 'param_A'
            settings = [['default', 0, 20], [2.0, 'fix'], [2.2, 'inf', 3.3]]
            instance.set_one(name, settings)
        """
        self.opt_params.set_one(name, settings)

        # update kim internal model param (note, set_one will update model_params)
        p_idx = self.model_params[name].index
        for c_idx, v in enumerate(self.model_params[name].value):
            self.kim_model.set_parameter(p_idx, c_idx, v)

        self.kim_model.clear_then_refresh()

        # reset influence distance in case it changes
        self.init_influence_distance()

    def update_model_params(self, params: Sequence[float]):
        """
        Update optimizing parameters (a sequence used by the optimizer) to the kim model.
        """
        # update from opt params to model params
        # TODO, in super().update_model_params(), we have parameter relation set,
        #   these parameters need to be updated here as well. However, in general
        #   we do not know how parameters are modified in parameter_relation,
        #   and it seems the only hope is to keep a copy of parameters and do some
        #   comparison to check which are modified and then set them.
        super().update_model_params(params)

        # update from model params to kim params
        n = self.get_num_opt_params()
        for i in range(n):
            _, value, p_idx, c_idx = self.get_opt_param_name_value_and_indices(i)
            self.kim_model.set_parameter(p_idx, c_idx, value)

        # refresh model
        self.kim_model.clear_then_refresh()

        if logger.getEffectiveLevel() == logging.DEBUG:
            params = self.get_kim_model_params()
            s = ""
            for name, p in params.items():
                s += f"\nname: {name}\n"
                s += str(p.as_dict())
            log_entry(logger, s, level="debug")

    def write_kim_model(self, path: Path = None):
        """
        Write out a KIM model that can be used directly with the kim-api.

        This function typically write two files to `path`: (1) CMakeLists.txt, and (2)
        a parameter file like A.model_params. `path` will be created if it does not exist.

        Args:
            path: Path to the a directory to store the model. If `None`, it is set to
                `./MODEL_NAME_kliff_trained`, where `MODEL_NAME` is the `model_name` that
                provided at the initialization of this class.

        Note:
            This only works for parameterized KIMModel models that support the writing of
            parameters.
        """
        present, required, error = self.kim_model.is_routine_present(
            kimpy.model_routine_name.WriteParameterizedModel
        )
        check_error(error, "kim_model.is_routine_is_routine_present")
        if not present:
            raise KIMModelError("This KIM model does not support writing parameters.")

        if path is None:
            model_name = self.model_name + "_kliff_trained"
            path = Path.cwd().joinpath(model_name)
        else:
            path = Path(path).expanduser().resolve()
            model_name = path.name

        if not path.exists():
            os.makedirs(path)

        path = str(path)
        model_name = str(model_name)

        error = self.kim_model.write_parameterized_model(path, model_name)
        check_error(error, "kim_model.write_parameterized_model")

        log_entry(logger, f"KLIFF trained model write to `{path}`", level="info")


class KIMModelError(Exception):
    def __init__(self, msg):
        super(KIMModelError, self).__init__(msg)
        self.msg = msg


def check_error(error, msg):
    if error != 0 and error is not None:
        msg = f"Calling `{msg}` failed.\nSee `kim.log` for more information."
        log_entry(logger, msg, level="error")
        raise KIMModelError(msg)


def report_error(msg):
    log_entry(logger, msg, level="error")
    raise KIMModelError(msg)



import os
import random
import tarfile
from collections.abc import Sequence
from pathlib import Path
from typing import Union

import numpy as np
import requests
import yaml


def length_equal(a, b):
    if isinstance(a, Sequence) and isinstance(b, Sequence):
        if len(a) == len(b):
            return True
        else:
            return False
    else:
        return True


def torch_available():
    try:
        import torch

        return True
    except ImportError:
        return False


def split_string(string: str, length=80, starter: str = None):
    r"""
    Insert `\n` into long string such that each line has size no more than `length`.

    Args:
        string: The string to split.
        length: Targeted length of the each line.
        starter: String to insert at the beginning of each line.
    """

    if starter is not None:
        target_end = length - len(starter) - 1
    else:
        target_end = length

    sub_string = []
    while string:
        end = target_end
        if len(string) > end:
            while end >= 0 and string[end] != " ":
                end -= 1
            end += 1
        sub = string[:end].strip()
        if starter is not None:
            sub = starter + " " + sub
        sub_string.append(sub)
        string = string[end:]

    return "\n".join(sub_string) + "\n"


def seed_all(seed=35, cudnn_benchmark=False, cudnn_deterministic=False):
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    if torch_available():
        import torch

        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)  # if using multi-GPU
        torch.backends.cudnn.benchmark = cudnn_benchmark
        torch.backends.cudnn.deterministic = cudnn_deterministic


def to_path(path: Union[str, Path]) -> Path:
    """
    Convert str (or filename) to pathlib.Path.
    """
    return Path(path).expanduser().resolve()


def download_dataset(dataset_name: str) -> Path:
    """
    Download dataset and untar it.

    Args:
        dataset_name: name of the dataset

    Returns:
        Path to the dataset
    """
    path = to_path(dataset_name)

    if not path.exists():

        tarball = path.with_suffix(".tar.gz")

        # download
        url = (
            f"https://raw.githubusercontent.com/mjwen/kliff/master/examples/"
            f"{dataset_name}.tar.gz"
        )

        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(tarball, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)

        # untar
        tf = tarfile.open(tarball, "r:gz")
        tf.extractall(path.parent)

        # # remove tarball
        # tarball.unlink()

    return path


def create_directory(path: Union[str, Path], is_directory=False):
    p = to_path(path)
    if is_directory:
        dirname = p
    else:
        dirname = p.parent
    if not dirname.exists():
        os.makedirs(dirname)


def yaml_dump(obj, filename):
    create_directory(filename)
    with open(to_path(filename), "w") as f:
        yaml.dump(obj, f, default_flow_style=False)


def yaml_load(filename):
    with open(to_path(filename), "r") as f:
        obj = yaml.safe_load(f)
    return obj



"""
Compute the root-mean-square error (RMSE) of a model prediction and reference values in
the dataset.
"""

from kliff.analyzers import EnergyForcesRMSE
from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.models import KIMModel
from kliff.utils import download_dataset

model = KIMModel(model_name="SW_StillingerWeber_1985_Si__MO_405512056662_005")

# load the trained model back
# model.load("kliff_model.yaml")


dataset_path = download_dataset(dataset_name="Si_training_set_4_configs")
tset = Dataset(dataset_path)
configs = tset.get_configs()

calc = Calculator(model)
calc.create(configs)

analyzer = EnergyForcesRMSE(calc)
analyzer.run(verbose=2, sort="energy")



import os
import sys

# TODO maybe move this to dataset


class Tree:
    def __init__(self):
        self.dirCount = 0
        self.stack_num_all = []
        self.stack_num_current = []
        self.string = ""
        self.started = False

    def count(self, absolute):
        if os.path.isfile(absolute) and absolute.endswith(".xyz"):
            return 1
        else:
            return 0

    def summary(self):
        s = "\n" + "=" * 80 + "\n"
        s += "                           KLIFF Dataset Count\n\n"
        s += 'Notation: "──dir_name (a/b)"\n'
        s += 'a: number of .xyz files in directory "dir_name"\n'
        s += 'b: number of .xyz files in directory "dir_name" and its subdirectories\n'
        s += "\n"
        s += self.string
        s += "\n" + "=" * 80 + "\n"

        return s

    def walk(self, directory, prefix=""):

        num_current = 0
        num_all = 0

        if not self.started:
            self.string += "{}{} (rpls_num_current_{}/rpls_num_all_{})\n".format(
                prefix, directory, self.dirCount, self.dirCount
            )
            self.stack_num_all.append("rpls_num_all_{}".format(self.dirCount))
            self.stack_num_current.append("rpls_num_current_{}".format(self.dirCount))
            self.dirCount += 1
            self.started = True

        filepaths = os.listdir(directory)
        for index, path in enumerate(filepaths):

            if path[0] == ".":
                continue

            absolute = os.path.join(directory, path)
            x = self.count(absolute)
            num_current += x
            num_all += x

            if index == len(filepaths) - 1:
                connector = "└──"
                prefix_new = prefix + "    "
            else:
                connector = "├──"
                prefix_new = prefix + "│   "

            if os.path.isdir(absolute):
                self.string += "{}{}{} (rpls_num_current_{}/rpls_num_all_{})\n".format(
                    prefix, connector, path, self.dirCount, self.dirCount
                )
                self.stack_num_all.append("rpls_num_all_{}".format(self.dirCount))
                self.stack_num_current.append(
                    "rpls_num_current_{}".format(self.dirCount)
                )
                self.dirCount += 1
                sub_current, sub_all = self.walk(absolute, prefix_new)
                num_all += sub_all

        self.string = self.string.replace(self.stack_num_all.pop(), str(num_all))
        self.string = self.string.replace(
            self.stack_num_current.pop(), str(num_current)
        )

        return num_current, num_all


def dataset_count(directory):
    tree = Tree()
    tree.walk(directory)
    print(tree.summary())


def split_dataset(source, target, nfold):
    if not os.path.exists(source):
        return 'input "{}" does not exists'.format(source)

    for i in range(nfold):
        path = os.path.join(target, "subset{}".format(i))
        if not os.path.exists(path):
            os.makedirs(path)

    return None


class Command:
    """
    Utility to manipulate dataset.
    """

    @staticmethod
    def add_arguments(parser):
        func = parser.add_argument
        func(
            "-c",
            "--count",
            type=str,
            metavar="directory",
            help='count the number of ".xyz" files in a given directory',
        )
        func(
            "-s",
            "--split",
            nargs=3,
            metavar=("<input>", "<output>", "<n folds>"),
            help="split dataset into multiple folds",
        )

    @staticmethod
    def run(args, parser):

        if args.count is not None:
            dataset_count(args.count)
        elif args.split is not None:
            source, target, nfold = args.split
            msg = split_dataset(source, target, int(nfold))
            if msg is not None:
                parser.error(msg)
        else:
            parser.print_help()


if __name__ == "__main__":

    directory = "."
    if len(sys.argv) > 1:
        directory = sys.argv[1]
    dataset_count(directory)



import logging
from typing import Callable, Dict, Optional

import numpy as np
from kliff.dataset.dataset import Configuration
from kliff.models.model import ComputeArguments, Model
from kliff.models.parameter import Parameter
from kliff.neighbor import NeighborList, assemble_forces, assemble_stress

logger = logging.getLogger(__name__)


class LJComputeArguments(ComputeArguments):
    """
    KLIFF built-in Lennard-Jones 6-12 potential computation functions.
    """

    implemented_property = ["energy", "forces", "stress"]

    def __init__(
        self,
        conf: Configuration,
        supported_species: Dict[str, int],
        influence_distance: float,
        compute_energy: bool = True,
        compute_forces: bool = True,
        compute_stress: bool = False,
    ):
        super(LJComputeArguments, self).__init__(
            conf,
            supported_species,
            influence_distance,
            compute_energy,
            compute_forces,
            compute_stress,
        )

        self.neigh = NeighborList(
            self.conf, influence_distance, padding_need_neigh=False
        )

    def compute(self, params: Dict[str, Parameter]):
        epsilon = params["epsilon"][0]
        sigma = params["sigma"][0]
        rcut = params["cutoff"][0]
        coords = self.conf.coords

        coords_including_padding = self.neigh.coords
        if self.compute_forces:
            forces_including_padding = np.zeros_like(coords_including_padding)

        energy = 0
        for i, xyz_i in enumerate(coords):
            neighlist, _, _ = self.neigh.get_neigh(i)
            for j in neighlist:
                xyz_j = coords_including_padding[j]
                rij = xyz_j - xyz_i
                r = np.linalg.norm(rij)
                if self.compute_forces:
                    phi, dphi = self.calc_phi_dphi(epsilon, sigma, r, rcut)
                    energy += 0.5 * phi
                    pair = 0.5 * dphi / r * rij
                    forces_including_padding[i] = forces_including_padding[i] + pair
                    forces_including_padding[j] = forces_including_padding[j] - pair
                elif self.compute_energy:
                    phi = self.calc_phi(epsilon, sigma, r, rcut)
                    energy += 0.5 * phi

        if self.compute_energy:
            self.results["energy"] = energy
        if self.compute_forces:
            forces = assemble_forces(
                forces_including_padding, len(coords), self.neigh.padding_image
            )
            self.results["forces"] = forces
        if self.compute_stress:
            volume = self.conf.get_volume()
            stress = assemble_stress(
                coords_including_padding, forces_including_padding, volume
            )
            self.results["stress"] = stress

    @staticmethod
    def calc_phi(epsilon, sigma, r, rcut):
        if r > rcut:
            phi = 0
        else:
            sor = sigma / r
            sor6 = sor * sor * sor
            sor6 = sor6 * sor6
            sor12 = sor6 * sor6
            phi = 4 * epsilon * (sor12 - sor6)
        return phi

    @staticmethod
    def calc_phi_dphi(epsilon, sigma, r, rcut):
        if r > rcut:
            phi = 0.0
            dphi = 0.0
        else:
            sor = sigma / r
            sor6 = sor * sor * sor
            sor6 = sor6 * sor6
            sor12 = sor6 * sor6
            phi = 4 * epsilon * (sor12 - sor6)
            dphi = 24 * epsilon * (-2 * sor12 + sor6) / r
        return phi, dphi


class LennardJones(Model):
    """
    KLIFF built-in Lennard-Jones 6-12 potential model.
    """

    def __init__(
        self, model_name="LJ6-12", params_relation_callback: Optional[Callable] = None
    ):
        super(LennardJones, self).__init__(model_name, params_relation_callback)

    def init_model_params(self):
        model_params = {
            "epsilon": Parameter(value=[1.0]),
            "sigma": Parameter(value=[2.0]),
            "cutoff": Parameter(value=[5.0]),
        }

        return model_params

    def init_influence_distance(self):
        return self.model_params["cutoff"][0]

    def init_supported_species(self):
        return None

    def get_compute_argument_class(self):
        return LJComputeArguments



from kliff.models.kim import KIMModel
from kliff.models.lennard_jones import LennardJones
from kliff.models.model import ComputeArguments, Model
from kliff.models.parameter import OptimizingParameters, Parameter
from kliff.utils import torch_available

__all__ = [
    "Parameter",
    "OptimizingParameters",
    "ComputeArguments",
    "Model",
    "LennardJones",
    "KIMModel",
]

if torch_available():
    from kliff.models.linear_regression import LinearRegression
    from kliff.models.model_torch import ModelTorch
    from kliff.models.neural_network import NeuralNetwork

    __all__.extend(["ModelTorch", "NeuralNetwork", "LinearRegression"])



import sys

from kliff.models import KIMModel


def get_avail_params(model_name):
    model = KIMModel(model_name)
    model.echo_model_params()


class Command:
    """
    Utility to manipulate dataset.
    """

    @staticmethod
    def add_arguments(parser):
        func = parser.add_argument
        func(
            "-a",
            "--avail-params",
            type=str,
            metavar="model_name",
            help="Get the available parameters of a KIM model to be used for fitting.",
        )

    @staticmethod
    def run(args, parser):

        if args.avail_params is not None:
            get_avail_params(args.avail_params)
        else:
            parser.print_help()


if __name__ == "__main__":

    model_name = "SW_StillingerWeber_1985_Si__MO_405512056662_005"
    if len(sys.argv) > 1:
        model_name = sys.argv[1]
    get_avail_params(model_name)
